"""Bit-blasting for signed integer arithmetic.

This module translates Boolean formulas that can
contain arithmetic expressions involving signed integers
to bitvector propositional formulas.

Reference
=========

Chapter 6, in particular pp. 158--159 from:
Kroening D. and Strichman O.
Decision Procedures, Springer
"""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from __future__ import absolute_import
import logging
import math

import networkx as nx

from omega.logic import lexyacc
from omega.logic import syntax as stx
from omega.logic.ast import Nodes as _Nodes


ALU_BITWIDTH = 32
DATA_TYPES = {'bool', 'int', 'saturating', 'modwrap'}
KEYS = {'type', 'dom', 'signed', 'width', 'bitnames'}
logger = logging.getLogger(__name__)


def bitblast(f, vrs, defs=None):
    """Flatten formula `f` to bitvector logic.

    @param f: quantified first-order action formula
    @type f: `str`
    @param vrs: symbol table of variables as returned by `bitblast_table`
    @type vrs: `dict`
    @param defs: operator definitions
    @type defs: `dict` that maps names (`str`) to expressions (`str`)
    """
    if defs is None:
        defs = dict()
    tree = _parser.parse(f)
    defs = {k: _parser.parse(v) for k, v in defs.items()}
    return tree.flatten(t=vrs, defs=defs)


def bitblast_table(table):
    """Return table of variables for bitvectors.

    `table` is a `dict` that maps each variable
    to a `dict` with attributes:

      - "type": `in ('bool', 'saturating', 'modwrap', 'int')`
      - "dom": `tuple([min, max])` where `min, max` are `int`
        used only if "type" is an integer
      - "init" (optional)
    """
    t = dict()
    for var, d in table.items():
        dtype = d['type']
        assert dtype in DATA_TYPES, (var, dtype)
        b = dict(d)  # cp other keys
        for k in KEYS:
            b.pop(k, None)
        if dtype == 'bool':
            b.update(type='bool')
        elif dtype in ('saturating', 'modwrap', 'int'):
            dom = d['dom']
            assert len(dom) == 2, dom
            signed, width = dom_to_width(dom)
            b.update(type='int',
                     signed=signed, width=width,
                     dom=dom)
        else:
            raise Exception(
                'unknown type: "{dtype}"'.format(dtype=dtype))
        if dtype == 'int':
            logger.info(
                '"int" found as type '
                '(instead of "saturating")')
        t[var] = b
    _check_data_types(t)
    _add_bitnames(t)
    logger.info('-- done bitblasting vars table\n')
    return t


def type_invariants(table):
    """Return type invariants for variables.

    @param table: bitblasted table of integers and Booleans
    @return: `(init, safety)`
    @rtype: `tuple` of `dict`,
        each `dict` maps variables to `list` of `str`.
    """
    init = dict()
    safety = dict()
    for var, d in table.items():
        init[var] = list()
        safety[var] = list()
        dtype = d['type']
        # initial value
        # imperative var or free var assigned at decl ?
        ival = d.get('init')
        if ival is not None:
            c = _init_to_logic(var, d)
            init[var].append(c)
        # ranged bitfield safety constraints
        if dtype == 'bool':
            continue
        dom = d['dom']
        # int var
        assert dtype in ('int', 'saturating', 'modwrap'), dtype
        dmin, dmax = dom
        # saturating semantics ?
        if dtype not in ('saturating', 'int'):
            # must range between powers of two
            v = abs(dmin) + 1
            assert not (v & (v - 1)), (var, dmin)
            v = abs(dmax) + 1
            assert not (v & (v - 1)), (var, dmax)
            continue
        # still included, for use with counters
        # during transducer construction,
        # because closure not taken for the counters
        s = '({min} <= {x})  /\  ({x} <= {max})'.format(
            min=dmin, max=dmax, x=var)
        init[var].append(s)
        # prime needed to enforce limits now, not one step later,
        # otherwise env can violate limits, if that will force
        # sys to lose in the next time step.
        s = (
            "    ({min} <= {x})  /\  ({x} <= {max}) "
            " /\ ({min} <= {x}') /\  ({x}' <= {max})").format(
                min=dmin, max=dmax, x=var)
        safety[var].append(s)
    return init, safety


def _init_to_logic(var, d):
    """Return logic formulas for initial condition."""
    if d['type'] == 'bool':
        op = '<=>'
    else:
        op = '='
    return '{var} {op} {value}'.format(
        op=op, var=var, value=d['init'])


def dom_to_width(dom):
    """Return whether integer variable is signed and its bit width.

    @param dom: the variable's range
    @type dom: `(MIN, MAX)` where `MIN, MAX` are integers
    """
    minval, maxval = dom
    logger.debug('int in ({m}, {M})'.format(
        m=minval, M=maxval))
    signed = (minval < 0) and (maxval >= 0)
    absval = max(abs(minval), abs(maxval))
    width = absval.bit_length()
    if width == 0:
        assert minval == maxval, (minval, maxval)
        # TODO: optimize by substituting values
        # for variables that are constants
        width = 1
    if signed:
        width = width + 1
    return signed, width


def _add_bitnames(t):
    """Map each integer to a list of bit variables."""
    for var, d in t.items():
        if d['type'] != 'int':
            continue
        assert d['type'] == 'int', d['type']
        if stx.isprimed(var):
            name = stx.unprime(var)
            prime = stx.PRIME
        else:
            name = var
            prime = ''
        bits = [
            '{name}_{i}{prime}'.format(
                name=name, i=i, prime=prime)
            for i in range(d['width'])]
        are_booleans = list(filter(t.__contains__, bits))
        assert not are_booleans, (bits, t)
        d['bitnames'] = bits


def _check_data_types(t):
    types = {'bool', 'int'}
    for var, d in t.items():
        if d['type'] in types:
            continue
        raise Exception(
            'unknown type: "{dtype}"'.format(dtype=d['type']))


def bit_table(variables, table):
    """Return symbol table of bits.

    For symbol table definition, see `bitblast_table`.

    @param variables: include only these variables
    @type variables: `set`
    @param table: symbol table of integer and Boolean variables
    @type table: `dict` of `dict`
    @return: symbol table of bits
    @rtype: `dict` of `dict`
    """
    dout = dict()
    keys = {'type', 'dom', 'signed',
            'width', 'bitnames', 'init'}
    for var in variables:
        d = table[var]
        dtype = d['type']
        if dtype == 'int':
            c = d['bitnames']
        elif dtype == 'bool':
            c = [var]
        else:
            raise Exception(
                'unknown type "{t}"'.format(t=dtype))
        dcp = dict(d)  # cp other keys
        for k in keys:
            dcp.pop(k, None)
        for bit in c:
            dbit = dict(dcp)
            dbit.update(type='bool')
            dout[bit] = dbit
    return dout


def list_bits(variables, table):
    """Collect bits of all (integer) `variables`."""
    s = set()
    for x in variables:
        # Boolean ?
        # if table[x]['type'] == 'bool':
        #     s.add(x)
        #     continue
        # integer
        bits = table[x]['bitnames']
        s.update(bits)
    return s


def map_bits_to_integers(table):
    """Return `dict` that maps each bit to an integer or Boolean."""
    bit2int = dict()
    for var, d in table.items():
        dtype = d['type']
        if dtype in ('int', 'saturating', 'modwrap'):
            assert 'bitnames' in d, d
            a = {b: var for b in d['bitnames']}
        elif dtype == 'bool':
            a = {var: var}
        else:
            raise Exception(
                'unknown var type "{dtype}"'.format(
                    dtype=dtype))
        bit2int.update(a)
    assert len(bit2int) >= len(table)
    return bit2int


def bitfield_to_int_states(g, t):
    """Convert bitfields to integers for "state" at each node.

    @type g: `networkx.DiGraph`
    @type t: `VariablesTable`

    @rtype: `networkx.Digraph`
    """
    h = nx.DiGraph()
    for u, d in g.nodes(data=True):
        bit_state = d['state']
        int_state = bitfields_to_ints(bit_state, t)
        h.add_node(u, state=int_state)
    for u, v in g.edges():
        h.add_edge(u, v)
    assert len(g) == len(h), (len(g), len(h))
    # remove deadends, where env looses
    s = {1}
    while s:
        s = {n for n in h if not h.succ.get(n)}
        h.remove_nodes_from(s)
    assert h or not g, 'No loop in given graph `g`.'
    logger.debug(
        ('converted bitfields to integers.\n'
         'Strategy with vertices:\n  {v}\n'
         'and edges:\n {e}\n').format(
            v='\n  '.join(str(x) for x in h.nodes(data=True)),
            e=h.edges()))
    return h


def bitfields_to_ints(bit_state, t):
    """Convert bits to integer for state `dict`.

    @param bit_state: assignment to all bits
    @type bit_state: `dict`
    @type t: `VariablesTable`
    """
    int_state = dict()
    for flatname, d in t.items():
        if d['type'] == 'bool':
            int_state[flatname] = bit_state[flatname]
            continue
        # this is an integer var
        bitnames = d['bitnames']
        bitvalues = [bit_state[b] for b in bitnames]
        _append_sign_bit(bitvalues, flatname, d)
        int_state[flatname] = twos_complement_to_int(bitvalues)
    return int_state


def make_table(d, env_vars=None):
    """Return symbol table from "simple" `dict`.

    @param env_vars: assign `owner` attribute to
        `'env'` if in this set, otherwise to `'sys'`.
    """
    if env_vars is None:
        env_vars = set()
    t = dict()
    for var, dom in d.items():
        if dom == 'bool':
            dtype = 'bool'
            dom = None
        else:
            assert isinstance(dom, tuple), (var, dom)
            assert len(dom) == 2, (var, dom)
            dtype = 'saturating'
        if var in env_vars:
            owner = 'env'
        else:
            owner = 'sys'
        t[var] = dict(type=dtype, dom=dom, owner=owner)
    return t


def make_symbol_table(vrs):
    """Return table of declarations from "simple" `dict`."""
    d = dict()
    for var, dom in vrs.items():
        if dom == 'bool':
            d[var] = dict(type='bool')
        else:
            assert len(dom) == 2, dom
            d[var] = dict(type='int', dom=dom)
    return d


def make_dummy_table():
    """Example of a symbol table."""
    t = dict(x=dict(type='bool'),
             y=dict(type='bool'),
             z=dict(type='int', signed=False, width=2),
             w=dict(type='int', signed=False, width=2))
    return t


class Nodes(_Nodes):
    """Return object with AST node classes as attributes."""

    opmap = {
        'false': '0', 'true': '1',
        '~': '!',
        r'\/': '|', '/\\': '&',
        '=>': '| !',
        '=>': '| !',
        '<=>': '! ^',
        'ite': 'ite', '@': '',
        '\A': '\A', '\E': '\E', '\S': '\S',
        'X': '',
        # 'G': '[]', 'F': '<>',
        '<': '<', '<=': '<=', '=': '=',
        '>=': '>=', '>': '>', '!=': '!=', '/=': '!=',
        '+': '+', '-': '-'}

    class Operator(_Nodes.Operator):
        def flatten(self, mem=None, *arg, **kw):
            if self.operator in ('\A', '\E'):
                assert mem is None, mem
                x, e = self.operands
                assert isinstance(x, list), x
                # list bits
                # t = kw['t']
                bits = list()
                for v in x:
                    flat = _flatten_var(v, mem=None, **kw)
                    bits.extend(flat)
                cube = (len(bits) - 1) * '& ' + ' '.join(bits)
                u = e.flatten(mem=None, *arg, **kw)
                r = ' {op} {cube} {u}'.format(
                    op=Nodes.opmap[self.operator],
                    cube=cube,
                    u=u)
                return r
            if self.operator == '\S':
                x, e = self.operands
                assert isinstance(x, list), x
                unique = set(new.value for new, old in x)
                assert len(unique) == len(x), x  # duplicates ?
                t = kw['t']
                rename = list()
                for new, old in x:
                    # take same values ?
                    old_var = old.value
                    new_var = new.value
                    d_old = t[old_var]
                    d_new = t[new_var]
                    assert d_old['type'] == d_new['type']
                    assert d_old.get('dom') == d_new.get('dom')
                    # flatten
                    a = _flatten_var(old, mem=mem, **kw)
                    b = _flatten_var(new, mem=mem, **kw)
                    assert len(a) == len(b), (a, b)
                    rename.extend(zip(a, b))
                pairs = ' '.join(
                    '{b} {a}'.format(a=a, b=b)
                    for a, b in rename)
                u = e.flatten(mem=mem, **kw)
                r = ' {op} ${n} {pairs} {u}'.format(
                    op=Nodes.opmap[self.operator],
                    n=2 * len(rename),
                    pairs=pairs,
                    u=u)
                return r
            if self.operator == 'LET':
                # populate local scope
                global_defs = kw.get('defs', dict())
                defs = dict(global_defs)  # isolate context
                local_defs, e = self.operands
                kw = dict(kw)  # local scope instead
                kw.pop('defs')
                for op_def in local_defs:
                    op_def.flatten(defs=defs, mem=mem, *arg, **kw)
                # flatten expr in local scope
                r = e.flatten(defs=defs, mem=mem, *arg, **kw)
                return r
            if self.operator == '@':
                # priming of state predicate BDDs unsupported yet
                assert not kw.get('prime'), kw
                x = int(self.operands[0].value)
                assert x != 0, x
                return '@' + str(x)
            if self.operator == '/\\':
                ops = [
                    u.flatten(mem=mem, *arg, **kw)
                    for u in self.operands]
                return stx.conj_prefix(ops)
            if self.operator == r'\/':
                ops = [
                    u.flatten(mem=mem, *arg, **kw)
                    for u in self.operands]
                return stx.disj_prefix(ops)
            if self.operator != 'ite':
                return super(Nodes.Operator, self).flatten(
                    mem=mem, *arg, **kw)
            # ternary conditional
            assert self.operator == 'ite', self.operator
            x = self.operands[0].flatten(mem=None, *arg, **kw)
            y = self.operands[1].flatten(mem=mem, *arg, **kw)
            z = self.operands[2].flatten(mem=mem, *arg, **kw)
            # ternary connective ?
            if mem is None:
                return ite_connective(x, y, z)
            else:
                p, q = equalize_width(y, z)
                r, ite_mem = ite_function(x, p, q, start=len(mem))
                mem.extend(ite_mem)
                return r

    class Unary(_Nodes.Unary):
        def flatten(self, *arg, **kw):
            logger.info('flatten "{s}"'.format(s=repr(self)))
            if self.operator == 'X':
                kw.update(prime=True)
                # avoid making it a string
                # (because in arithmetic context it is a list)
                return self.operands[0].flatten(*arg, **kw)
            return ' {op} {x}'.format(
                op=Nodes.opmap[self.operator],
                x=self.operands[0].flatten(*arg, **kw))

    # prefix and rm parentheses
    class Binary(_Nodes.Binary):
        def flatten(self, *arg, **kw):
            """Prefix flattener."""
            logger.info('flatten "{s}"'.format(s=repr(self)))
            op = self.operator
            x, y = self.operands
            if op == '==':
                name = x.value
                defs = kw['defs']  # must be in some context
                assert name not in defs, (
                    'attempted redefining "{name}"'.format(
                        name=name))
                defs[name] = y
                return
            x = x.flatten(*arg, **kw)
            y = y.flatten(*arg, **kw)
            if op == '..':
                return (x, y)
            elif op == r'\in':
                assert len(y) == 2, y
                e = x
                a, b = y
                x = flatten_comparator('<=', a, e, mem=list())
                y = flatten_comparator('<=', e, b, mem=list())
                op = '/\\'
            op = Nodes.opmap[op]
            assert isinstance(x, str), x
            assert isinstance(y, str), y
            return ' {op} {x} {y} '.format(
                op=op, x=x, y=y)

    class Var(_Nodes.Var):
        def flatten(self, prime=None, mem=None,
                    t=None, defs=None, *arg, **kw):
            logger.info('flatten "{s}"'.format(s=repr(self)))
            name = self.value
            # operator definition ?
            if defs is not None and name in defs:
                u = defs[name]
                # BDD ?
                if hasattr(u, 'var'):
                    return str(u)
                # list at arithmetic level ?
                if isinstance(u, list):
                    return u
                # ast
                return u.flatten(
                    prime=prime, mem=mem, t=t, defs=defs,
                    *arg, **kw)
            if _is_bool_var(name, t):
                # Boolean scope ?
                if name in t:
                    assert mem is None, mem
                return '{v}{prime}'.format(
                    v=name, prime=stx.PRIME if prime else '')
            assert name in t, (
                '"{name}" neither var nor operator'.format(
                    name=name))
            # arithmetic context
            # must be integer variable
            # a refinement occurs here automatically
            bits = var_to_twos_complement(name, t)
            bits = ["{b}{prime}".format(
                    b=b, prime=stx.PRIME
                    if not b[0].isdigit() and prime else '')
                for b in bits]
            return bits

    class Num(_Nodes.Num):
        def flatten(self, *arg, **kw):
            logger.info('flatten "{s}"'.format(s=repr(self)))
            return int_to_twos_complement(self.value)

    class Bool(_Nodes.Bool):
        def flatten(self, *arg, **kw):
            return Nodes.opmap[self.value.lower()]

    class Comparator(_Nodes.Comparator):
        def flatten(self, mem=None, *arg, **kw):
            logger.info('flatten "{s}"'.format(s=repr(self)))
            assert mem is None, (
                '"{expr}" appears in arithmetic scope'.format(
                    expr=self))
            mem = list()
            p = self.operands[0].flatten(mem=mem, *arg, **kw)
            q = self.operands[1].flatten(mem=mem, *arg, **kw)
            return flatten_comparator(self.operator, p, q, mem)

    class Arithmetic(_Nodes.Arithmetic):
        def flatten(self, mem=None, *arg, **kw):
            if self.operator == '<<>>':
                return flatten_truncator(
                    self.operands, mem=mem, *arg, **kw)
            logger.info('flatten "{s}"'.format(s=repr(self)))
            # only for testing purposes
            assert mem is not None, (
                'Arithmetic formula "{f}" '
                'in Boolean scope.').format(
                    f=self)
            p = self.operands[0].flatten(mem=mem, *arg, **kw)
            q = self.operands[1].flatten(mem=mem, *arg, **kw)
            return flatten_arithmetic(self.operator, p, q, mem)


_parser = lexyacc.Parser(nodes=Nodes())


def flatten_truncator(operands, mem=None, *arg, **kw):
    """Return integer truncated to given width."""
    logger.info(
        '++ flatten truncator "{s}"'.format(s=operands))
    p = operands[0].flatten(mem=mem, *arg, **kw)
    assert isinstance(p, list), p
    y = operands[1]
    assert isinstance(y, Nodes.Num), (type(y), y)
    n = int(y.value)
    tr = truncate(p, n)
    # if extended, should not use MSB of truncation
    tr.append('0')
    logger.debug('truncation result: {tr}\n'.format(tr=tr))
    logger.info('-- done flattening truncator.\n')
    return tr


def flatten_comparator(operator, x, y, mem):
    """Return flattened comparator formula."""
    logger.info(
        '++ flatten comparator "{op}" ...'.format(op=operator))
    assert isinstance(x, list), x
    assert isinstance(y, list), y
    p, q = equalize_width(x, y)
    assert len(p) == len(q), (p, q)
    logger.debug('p = {p}\nq = {q}'.format(p=p, q=q))
    negate = False
    if operator in {'=', '!=', '/='}:
        r = inequality(p, q, mem)
        if operator == '=':
            negate = True
        else:
            assert operator == '!=' or operator == '/=', operator
    elif operator in {'<', '<=', '>=', '>'}:
        swap = False
        if operator == '<=':
            negate = True
            swap = True
        elif operator == '>':
            swap = True
        elif operator == '>=':
            negate = True
        else:
            assert operator == '<', operator
        if swap:
            p, q = q, p
        r = less_than(p, q, mem)
    else:
        raise ValueError(
            'unknown operator "{op}"'.format(op=operator))
    if negate:
        r = '! {r}'.format(r=r)
    mem.append(r)
    logger.debug('mem = {mem}'.format(mem=_format_mem(mem=mem)))
    logger.debug('-- done flattening "{op}"\n'.format(op=operator))
    return '$ {n} {s}'.format(n=len(mem), s=' '.join(mem))


def inequality(p, q, mem):
    """Return bitvector propositional formula for '/='."""
    assert len(p) == len(q), (p, q)
    return ' '.join('| ^ {a} {b}'.format(a=a, b=b)
                    for a, b in zip(p, q)) + ' 0'


def less_than(p, q, mem):
    """Return bitvector propositional formula for '<'."""
    res, res_mem, carry = adder_subtractor(
        p, q, add=False, start=len(mem))
    mem.extend(res_mem)
    return '^ ! ^ {a} {b} {carry}'.format(
        a=p[-1], b=q[-1], carry=carry)


def flatten_arithmetic(operator, p, q, mem):
    """Return flattened arithmetic expression."""
    logger.info(
        '++ flatten arithmetic operator "{op}"'.format(
            op=operator))
    assert isinstance(p, list), p
    assert isinstance(q, list), q
    assert isinstance(mem, list), mem
    start = len(mem)
    if operator in {'+', '-'}:
        add = (operator == '+')
        result, res_mem, _ = adder_subtractor(p, q, add, start)
    elif operator == '*':
        result, res_mem = multiplier(p, q, start)
    elif operator == '/':
        result, _, res_mem = restoring_divider(p, q, start)
    elif operator == '%':
        _, result, res_mem = restoring_divider(p, q, start)
    else:
        raise ValueError(
            'Unknown arithmetic operator "{op}"'.format(
                op=operator))
    mem.extend(res_mem)
    logger.info('-- done flattening "{op}"\n'.format(op=operator))
    return result


def restoring_divider(x, y, start=0):
    """Return divider for bitvectors `x`, `y`.

    @param x: dividend
    @param y: divisor
    @type x, y: `list`
    @param start: memory address to start indexing from
    @type start: `int` >= 0

    @return: (quotient, remainder, memory)
    @rtype: `tuple(list, list, list)`
    """
    # TODO: propagate to propositional context
    # constraint that detects zero divisor
    assert start >= 0, start
    assert isinstance(x, list), x
    assert isinstance(y, list), y
    mem = list()
    j = start
    # rectify
    a, a_mem = abs_(x, j)
    j = _extend_memory(mem, a_mem, j)
    b, b_mem = abs_(y, j)
    j = _extend_memory(mem, b_mem, j)
    # divide
    quo, rem, div_mem = _restoring_divider(a, b, start=j)
    j = _extend_memory(mem, div_mem, j)
    # fix signs
    x_sign = sign(x)
    y_sign = sign(y)
    opposite_signs = '^ {x} {y}'.format(x=x_sign, y=y_sign)
    quo, neg_mem = _negate_if(opposite_signs, quo, start=j)
    j = _extend_memory(mem, neg_mem, j)
    rem, neg_mem = _negate_if(x_sign, rem, start=j)
    j = _extend_memory(mem, neg_mem, j)
    return quo, rem, mem


def _restoring_divider(x, y, s=None, start=0):
    """Return stage `s` of divider (positive).

    @param x: dividend
    @param y: divider
    @type x, y: positive numbers in two's complement
        `list`
    @param s: desired stage of divider
    @type s: `int` with: `-1 <= s <= len(x)`
    @param start: memory address to start indexing from
    @type start: `int` >= 0
    """
    assert start >= 0, start
    assert isinstance(x, list), x
    assert isinstance(y, list), y
    n = len(x)
    # init
    if s is None:
        # double widths
        y = pad(y, 2 * n)
        y = fixed_shift(y, n, left=True)
        s = n - 1
    assert -1 <= s < n, (s, n)
    mem = list()
    j = start
    # base stage: -1
    if s == -1:
        quo = list()
        rem = pad(x, 2 * n)
        return quo, rem, mem
    # stages: 0 ... n - 1
    # recurse
    quo, p, div_mem = _restoring_divider(x, y, s - 1, start=j)
    j = _extend_memory(mem, div_mem, j)
    # this stage
    # diff
    shifted_p = fixed_shift(p, 1, left=True)
    r, sum_mem, carry = adder_subtractor(shifted_p, y, add=False,
                                         start=j, extend_by=0)
    j = _extend_memory(mem, sum_mem, j)
    # quotient
    sgn = sign(r)
    q = '! {sgn}'.format(sgn=sgn)
    quo.insert(0, q)
    # partial remainder
    rem, ite_mem = ite_function(q, r, shifted_p, start=j)
    j = _extend_memory(mem, ite_mem, j)
    # top stage ?
    if s == n - 1:
        rem = rem[n:]
    return quo, rem, mem


def multiplier(x, y, start=0):
    """Return the signed product of `x` and `y`.

    @param x, y: multiplicands
    @type x, y: `list`

    @return: (result, memory)
    @rtype: `tuple(list, list)`
    """
    assert isinstance(x, list), x
    assert isinstance(y, list), y
    nx = len(x)
    ny = len(y)
    n = nx + ny
    p, q = equalize_width(x, y, extend_by=min(nx, ny))
    res, mem = _multiplier(p, q, s=None, start=start)
    assert len(res) == n, (len(res), n)
    if n > ALU_BITWIDTH:
        print('WARNING: (bitvector) '
              'Truncating multiplication to {alu} bits.'.format(
                  alu=ALU_BITWIDTH))
        res = truncate(res, ALU_BITWIDTH)
    return res, mem


def _multiplier(x, y, s=None, start=0):
    """Return stage `s` of multiplier.

    @param x, y: multiplicands (in two's complement)
    @param s: desired stage of multiplier
    @type s: `int` with: `-1 <= s < len(y)`
    @param start: memory address to start indexing from
    @type start: `int` >= 0

    @return: (result, memory)
    @rtype: `tuple(list, list)`
    """
    assert start >= 0, start
    assert len(x) == len(y), (x, y)
    n = len(y)
    if s is None:
        s = n - 1
    assert -1 <= s < n, (s, n)
    # base stage: -1
    if s == -1:
        mem = list()
        return ['0'] * len(x), mem
    # stages: 0 ... n - 1
    # recurse
    j = start
    mul_res, mem = _multiplier(x, y, s=s - 1, start=j)
    j += len(mem)
    # this stage
    shifted_x = fixed_shift(x, s, left=True)
    z = ['& {a} {b}'.format(a=a, b=y[s]) for a in shifted_x]
    res, sum_mem, carry = adder_subtractor(
        mul_res, z, add=True, start=j, extend_by=0)
    j = _extend_memory(mem, sum_mem, j)
    assert len(res) == len(x), (x, res, mem)
    return res, mem


def adder_subtractor(x, y, add=True, start=0, extend_by=1):
    """Return sum of `p` and `q`, w/o truncation.

    Implements a ripple-carry adder-subtractor.
    The control signal is `add`.

    Reference
    =========
    https://en.wikipedia.org/wiki/Adder%E2%80%93subtractor
    https://en.wikipedia.org/wiki/Adder_%28electronics%29

    @param x, y: summands (in two's complement)
    @type x, y: `list` of bits
    @param add: if `True` then add, otherwise subtract
    @type add: `bool`
    @param start: insert first element at
        this index in memory structure
    @type start: `int` >= 0
    @param extend_by: extra sign-extension by so many bits
    @type extend_by: `int` >= 0

    @return: (result, memory, carry)
    @type: `tuple(list, list, str)`
    """
    assert start >= 0, start
    assert extend_by >= 0, extend_by
    assert isinstance(x, list), x
    assert isinstance(y, list), y
    dowhat = 'add' if add else 'subtract'
    logger.info('++ {what}...'.format(what=dowhat))
    p, q = equalize_width(x, y, extend_by=extend_by)
    assert len(p) == len(q), (p, q)
    logger.debug('p = {p}\nq = {q}'.format(p=p, q=q))
    # invert
    if add:
        carry = '0'
    else:
        q = ['! {bit}'.format(bit=b) for b in q]
        carry = '1'
    mem = list()
    result = list()
    # use a loop to emphasize the relation
    # between mem, result, carries
    for i, (a, b) in enumerate(zip(p, q)):
        k = start + 2 * i
        r = k + 1
        # full-adder
        # result
        mem.append('^ ^ {a} {b} {c}'.format(a=a, b=b, c=carry))
        result.append('? {k}'.format(k=k))
        # carry
        mem.append(
            '| & {a} {b} & ^ {a} {b} {c}'.format(a=a, b=b, c=carry))
        carry = '? {r}'.format(r=r)
    assert len(mem) == 2 * len(result), (mem, result)
    logger.debug('mem = {mem}\nres = {res}'.format(
        mem=_format_mem(mem), res=result))
    logger.info('-- done {what}ing\n'.format(what=dowhat))
    return result, mem, carry


def barrel_shifter(x, y, s=None, start=0):
    """Return left or right shift formula.

    Recursive implementatin of barrel shifter.
    Note that the shift distance must be represented as unsigned.

    @param x: shift (vector that is to be shifted)
    @type x: `list` of `str`
    @param y: shift distance
    @type y: `list` of `str` with `len(y) == math.log(len(x), 2)`
    @param s: desired stage of barrel shifter
    @type s: `int` with: `-1 <= s < len(y)`
    @param start: memory address to start indexing from
    @type start: `int` >= 0

    @return: 2-tuple:
      1. elements of composition of first `s` stages
      2. memory contents from stage 0 to stage `s`
    @rtype: `tuple([list, list])`
    """
    assert len(y) == math.log(len(x), 2), (x, y)
    if s is None:
        s = len(y) - 1
    assert -1 <= s < len(y), (s, y)
    assert start >= 0, start
    # base stage: -1
    if s == -1:
        mem = list()
        return x, mem
    # stages: 0 ... n - 1
    r, mem = barrel_shifter(x, y, s - 1, start=start)
    for i, a in enumerate(x):
        b = y[s]
        g = r[i]
        h = r[i - 2**s]
        if i < 2**s:
            z = '& ! {b} {g}'.format(b=b, g=g)
        else:
            z = '| & ! {b} {g} & {b} {h}'.format(b=b, g=g, h=h)
        mem.append(z)
    n = len(x)
    m = len(mem) - n
    c = ['? {i}'.format(i=i + m) for i in range(n)]
    assert len(c) == len(x), (c, x)
    return c, mem


def fixed_shift(x, c, left, logical=False, truncate=True):
    """Shift `x` by fixed distance `s`.

    @param x: shift (vector to be shifted)
    @type x: `list` of `str`
    @param c: shift distance
    @type c: `int` with: `0 <= c <= len(x)`
    @param left: if `True` shift left, else right
    @param logical: if `True` insert zeros,
        else replicate sign bit of `x`.
    @param truncate: if `True`, result has same width as `x`

    @return: shifted `x`
    @rtype: `list` of `str`
    """
    n = len(x)
    assert 0 <= c <= n, (c, n, x)
    if left:
        if truncate:
            return c * ['0'] + x[:n - c]
        else:
            return c * ['0'] + x
    # right shift
    # logical or arithmetic ?
    if logical:
        s = '0'
    else:
        s = x[-1]
    return x[c:] + c * [s]


def truncate(x, n):
    """Return first `n` bits of bitvector `x`.

    @type x: `list`
    @type n: `int` >= 0

    @rtype: `list`
    """
    assert n >= 0, n
    return x[:n]


def ite_function(a, b, c, start):
    """Return memory buffer elements for ite between integers.

    @param a: propositinal formula
    @type a: `str`
    @param b, c: arithmetic formula
    @type b, c: `list`
    @param start: continue indexing buffer cells from this value
    @type start: `int`

    @rtype: `list`
    """
    assert isinstance(a, str), a
    assert isinstance(b, list), b
    assert isinstance(c, list), c
    assert len(b) == len(c), (b, c)
    m = list()
    m.append(a)
    for p, q in zip(b, c):
        s = '| & {p} ? {i} & {q} ! ? {i}'.format(p=p, q=q, i=start)
        m.append(s)
    r = ['? {i}'.format(i=i + start + 1) for i, _ in enumerate(b)]
    return r, m


def ite_connective(a, b, c):
    """Return memory buffer for ternary conditional operator.

    Note that economizes by avoiding rewrite formulas.
    In Boolean context, the arguments a, b, c will always
    be variables of type bit, or Boolean constants,
    or the result of expressions as a memory buffer.

    @rtype: `str`
    """
    assert isinstance(a, str), a
    assert isinstance(b, str), b
    assert isinstance(c, str), c
    # local memory buffer
    return '$ 2 {a} | & {b} ? {i} & {c} ! ? {i}'.format(
        a=a, b=b, c=c, i=0)


def var_to_twos_complement(var, t):
    """Return `list` of bits in two's complement."""
    # little-endian encoding
    logger.info(
        '++ encode variable "{var}" to 2s complement'.format(
            var=var))
    _assert_var_in_table(var, t)
    d = t[var]
    # arithmetic operators defined only for integers
    if d['type'] == 'bool':
        raise TypeError((
            '2s complement undefined for '
            'Boolean variable "{var}"').format(var=var))
    bits = list(d['bitnames'])
    _append_sign_bit(bits, var, d)
    assert len(bits) > 1, bits
    logger.debug('encoded variable "{var}":\n\t{bits}'.format(
        var=var, bits=bits))
    logger.info('-- done encoding variable "{var}".\n'.format(
        var=var))
    return bits


def _append_sign_bit(bits, var, d):
    """Convert trimmed bitfield to two's complement.

    The bitfield `bits` is modified by appending a sign bit.
    The integer variable represented by
    `bits` should be sign-definite.
    As given, `bits` is a bitfield that
    stores the two's complement
    with omitted sign bit, because it is constant.

    @type bits: `list`
    @param d: attributes of integer
    @type d: `dict`
    """
    logger.debug('bits of "{var}": {bits}"'.format(
        var=var, bits=bits))
    # variable sign ?
    if d['signed']:
        logger.debug('variable "{var}" is signed'.format(var=var))
        if len(bits) < 2:
            raise ValueError((
                'A signed bitvector must have at least 2 bits.\n'
                'Got instead, for variable "{var}",'
                ' the bitvector:\n\t {bits}').format(
                    var=var, bits=bits))
        return
    # sign-definite
    logger.debug('variable "{var}" has fixed sign'.format(var=var))
    min_, max_ = d['dom']
    assert min_ * max_ >= 0, (min_, max_)
    # positive ?
    if min_ >= 0:
        sign_bit = '0'
    else:
        assert max_ < 0, max_  # only negative values
        sign_bit = '1'
    bits.append(sign_bit)
    assert len(bits) > 1, len(bits)


def _is_bool_var(name, t):
    # fol var ?
    if name in t:
        d = t[name]
        return (d['type'] == 'bool')
    else:
        # bit in a bitvector ?
        int_var = name.rsplit('_', 1)[0]
        is_bit = (
            int_var in t and
            name in t[int_var]['bitnames'])
        if is_bit:
            return True


def _assert_var_in_table(name, t):
    """Raise `Exception` if variable `name` not in table `t`."""
    # not in variables table ?
    if name in t:
        return
    raise Exception(
        'Variable "{name}" missing from symbol table:\n{t}'.format(
            name=name, t=t))


def int_to_twos_complement(s):
    """Return two's complement of `s` as `list` of `str`.

    @type s: such that `int(s)` is well-defined
    """
    logger.info(
        '++ convert integer "{s}" to 2s complement'.format(s=s))
    x = int(s)
    n = x.bit_length()
    if x >= 0:
        sign_bit = '0'
        y = x
    else:
        sign_bit = '1'
        y = 2**n + x
    m = max(n, 1)  # if y == 0 then n == 0
    bits = bin(y).lstrip('-0b').zfill(m)
    bits = list(reversed(bits))
    bits.append(sign_bit)
    x_ = twos_complement_to_int(bits)
    assert x == x_, (x, x_)
    assert len(bits) > 1, bits
    # bits = sign_extension(bits, n + 1)
    logger.info("two's complement of \"{s}\" is:\n\t{bits}".format(
        s=s, bits=bits))
    logger.info('-- done encoding int\n')
    return bits


def twos_complement_to_int(bits):
    """Return `int` equal to value of two's complement in `bits`."""
    n = len(bits)  # width including sign bit
    n = n - 1  # width w/o sign bit
    r = [int(b) for b in bits]
    return -r[-1] * 2**n + sum(
        b * 2**i
        for i, b in enumerate(r[:-1]))


def abs_(x, start=0):
    """Return absolute value of `x`.

    @param start: memory address to start indexing from
    @type start: `int` >= 0
    """
    return _negate_if(sign(x), x, start)


def _negate_if(guard, x, start=0):
    """Return conditional negation of `x`.

    @param guard: if `True`, then negate
    @type guard: `str`
    """
    assert isinstance(x, list), x
    assert start >= 0, start
    n = len(x)
    zero = pad(['0'], len(x))
    j = start
    neg_x, mem, _ = adder_subtractor(
        zero, x, add=False, start=j, extend_by=1)
    ext_x = sign_extension(x, n + 1)
    j += len(mem)
    r, ite_mem = ite_function(guard, neg_x, ext_x, start=j)
    j = _extend_memory(mem, ite_mem, j)
    return r, mem


def equalize_width(x, y, extend_by=0):
    """Return bitvectors of equal len by applying sign extension."""
    logger.info('++ equalize width...')
    logger.debug(
        'before equalization:\n\t x = {x}\n\t y = {y}'.format(
            x=x, y=y))
    n = max(len(x), len(y))
    n += extend_by
    p = sign_extension(x, n)
    q = sign_extension(y, n)
    logger.debug('after extension:\n\t x = {p}\n\t y = {q}'.format(
        p=p, q=q))
    assert len(p) == len(q), (p, q)
    assert len(p) == n, (p, n)
    logger.info('-- done equalizing.\n')
    return p, q


def sign_extension(x, n):
    """Extend two's complement of `x` to `n` bits width.

    Note that the width of two's complement should be at
    least 2 bits, otherwise it doesn't make sense.

    Reference
    =========
    https://en.wikipedia.org/wiki/Sign_extension

    @type x: `list` of `str`
    @type n: `int` with: `len(x) <= n < 32`
    """
    logger.debug(
        '++ sign extension to {n} bits of: {x}'.format(x=x, n=n))
    assert isinstance(x, list), x
    assert n < ALU_BITWIDTH, n
    m = len(x)
    if m < 2:
        raise ValueError('"{x}" has less than 2 bits.'.format(x=x))
    if n < m:
        raise ValueError(
            'Extension width is {n} < {m} = len(x)'.format(n=n, m=m))
    y = x + (n - m) * [x[-1]]
    assert y[:len(x)] == x, (y, x)
    assert len(y) == n, (y, n)
    logger.debug('-- result of extension: {y}\n'.format(y=y))
    return y


def pad(x, n):
    """Return `x` left-padded with zeros to width `n`.

    @type x: `list`
    @type n: `int` > 0
    """
    m = n - len(x)
    assert m > 0, (n, m, x)
    return x + m * ['0']


def sign(x):
    """Return sign bit (MSB) of `x` in two's complement.

    @type x: `list` of `str`
    @return: `str` ('0' or '1')
    """
    return x[-1]


def _extend_memory(mem, more_mem, start):
    """Return new start address after appending memory.

    @param mem: memory to be extended
    @param more_mem: memory elements to append
    @type mem, more_mem: `list`
    @param start: last `mem` element has address `start - 1`
    @type start: `int` with `start >= len(mem)`
    """
    assert start >= len(mem), (start, len(mem))
    mem.extend(more_mem)
    start += len(more_mem)
    return start


def _format_mem(mem):
    """Return pretty string for printing memory contents."""
    return 'memory:\n{mem}\n'.format(
        mem='\n'.join('{i}: {v}'.format(i=i, v=v)
                      for i, v in enumerate(mem)))


def _flatten_var(v, *arg, **kw):
    """Return `list` of bits, for both integer and Boolean var."""
    flat = v.flatten(*arg, **kw)
    # bool ?
    if stx.isinstance_str(flat):
        flat = [flat]
    bits = _filter_trailing_zeros(flat)
    return bits


def _filter_trailing_zeros(flat):
    return [b for b in flat if not b[0].isdigit()]



"""Tests for `omega.games.gr1`."""
import logging
import pprint

from dd import mdd
from nose.tools import assert_raises

from omega.symbolic import enumeration
from omega.symbolic import symbolic
from omega.symbolic import temporal as trl
from omega.games import gr1


log = logging.getLogger('astutils')
log.setLevel('ERROR')
log = logging.getLogger('omega')
log.setLevel('ERROR')


def test_streett_trivial_loop():
    a = trl.default_streett_automaton()
    a.declare_variables(x='bool')
    a.varlist['sys'] = ['x']
    # solve
    assert len(a.win['<>[]']) == 1
    assert len(a.win['[]<>']) == 1
    z, yij, xijk = gr1.solve_streett_game(a)
    assert z == a.bdd.true, z
    # transducer
    gr1.make_streett_transducer(z, yij, xijk, a)
    # vars
    assert '_goal' in a.vars, a.vars
    vt = a.vars['_goal']['type']
    assert vt == 'int', vt
    dom = a.vars['_goal']['dom']
    assert dom == (0, 0), dom
    assert 'x' in a.vars, a.vars
    # init
    init = a.init['impl_env']
    init_ = a.add_expr('_goal = 0')
    assert init == init_, a.bdd.to_expr(init)
    # action
    action = a.action['impl']
    s = "(_goal = 0) /\ (_goal' = 0)"
    action_ = a.add_expr(s)
    assert action == action_, a.bdd.to_expr(action)


def test_rabin_trivial_loop():
    a = trl.default_rabin_automaton()
    a.acceptance = 'Rabin(1)'
    a.declare_variables(x='bool')
    a.varlist['sys'] = ['x']
    # solve
    assert len(a.win['<>[]']) == 1
    assert len(a.win['[]<>']) == 1
    zk, yki, xkijr = gr1.solve_rabin_game(a)
    assert zk[-1] == a.bdd.true, zk
    # transducer
    gr1.make_rabin_transducer(zk, yki, xkijr, a)
    # vars
    assert '_goal' in a.vars, a.vars
    vt = a.vars['_goal']['type']
    assert vt == 'int', vt
    dom = a.vars['_goal']['dom']
    assert dom == (0, 0), dom
    assert '_hold' in a.vars, a.vars
    vt = a.vars['_hold']['type']
    assert vt == 'int', vt
    dom = a.vars['_hold']['dom']
    assert dom == (0, 1), dom
    assert 'x' in a.vars, a.vars
    # init
    init = a.init['impl_env']
    init_ = a.add_expr('(_goal = 0) /\ (_hold = 1)')
    assert init == init_, a.bdd.to_expr(init)
    # action
    action = a.action['impl']
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "(_hold' = 0)")
    action_ = a.add_expr(s)
    assert action == action_, a.bdd.to_expr(action)


def test_streett_deadend():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x=(0, 10))
    aut.varlist['sys'] = ['x']
    aut.action['sys'] = aut.add_expr("(x = 0) /\ (x' = 5)")
    z, _, _ = gr1.solve_streett_game(aut)
    win_set = z
    assert win_set == aut.bdd.false, win_set


def test_rabin_deadend():
    aut = trl.default_streett_automaton()
    aut.acceptance = 'Rabin(1)'
    aut.declare_variables(x=(0, 101))
    aut.varlist['sys'] = ['x']
    aut.action['sys'] = aut.add_expr("(x = 1) /\ (x' = 96 - x)")
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.false, win_set


def test_streett_always_x():
    # always x
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['sys'] = ['x']
    aut.moore = False
    aut.action['sys'] = aut.add_expr("x' ")
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.true, z
    # transducer
    gr1.make_streett_transducer(z, yij, xijk, aut)
    # init
    init = aut.init['impl_env']
    init_ = aut.add_expr('_goal = 0')
    assert init == init_, aut.bdd.to_expr(init)
    # action
    action = aut.action['impl']
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "x' ")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)
    #
    # always ~ x
    aut.init['env'] = aut.add_expr(' ~ x')
    aut.action['sys'] = aut.add_expr(' ~ x')
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.add_expr(' ~ x'), aut.bdd.to_expr(z)
    # tranducer
    gr1.make_streett_transducer(z, yij, xijk, aut)
    assert action_refined(aut)
    init = aut.init['impl_env']
    init_ = aut.add_expr('(_goal = 0) /\ ~ x')
    assert init == init_, aut.bdd.to_expr(init)
    action = aut.action['impl']
    s = "(_goal = 0) /\ (_goal' = 0) /\ ~ x /\ ~ x'"
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)


def test_rabin_always_x():
    # always x
    aut = trl.default_rabin_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['sys'] = ['x']
    aut.acceptance = 'Rabin(1)'
    aut.action['sys'] = aut.add_expr("x' ")
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, win_set
    # transducer
    gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    # init
    init = aut.init['impl_env']
    s = '(_goal = 0) /\ (_hold = 1)'
    init_ = aut.add_expr(s)
    assert init == init_, aut.bdd.to_expr(init)
    # action
    action = aut.action['impl']
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "(_hold' = 0) /\ "
        "x' ")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)
    #
    # always ~ x
    aut.init['env'] = aut.add_expr(' ~ x')
    aut.action['sys'] = aut.add_expr(' ~ x')
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    win_set_ = aut.add_expr(' ~ x')
    assert win_set == win_set_, aut.bdd.to_expr(win_set)
    # transducer
    gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut)
    init = aut.init['impl_env']
    s = (
        '(_goal = 0) /\ (_hold = 1) /\ '
        ' ~ x')
    init_ = aut.add_expr(s)
    assert init == init_, aut.bdd.to_expr(init)
    action = aut.action['impl']
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "(_hold' = 0) /\ "
        " ~ x /\ ~ x'")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)


def test_streett_counter():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['sys'] = ['x']
    aut.action['sys'] = aut.add_expr("x => ~ x' ")
    aut.win['[]<>'] = [aut.add_expr('x')]
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.true, z
    # transducer
    gr1.make_streett_transducer(z, yij, xijk, aut)
    init = aut.init['impl_env']
    init_ = aut.add_expr('_goal = 0')
    assert init == init_, aut.bdd.to_expr(init)
    action = aut.action['impl']
    assert action_refined(aut)
    # regression
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "(x <=> ~ x')")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)


def test_rabin_counter():
    aut = trl.default_rabin_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['sys'] = ['x']
    aut.plus_one = False
    aut.qinit = '\A \A'
    aut.action['sys'] = aut.add_expr("x => ~ x' ")
    aut.win['[]<>'] = [aut.add_expr('x')]
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, aut.bdd.to_expr(win_set)
    # transducer
    gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut)
    init = aut.init['impl_env']
    s = '(_goal = 0) /\ (_hold = 1)'
    init_ = aut.add_expr(s)
    assert init == init_, aut.bdd.to_expr(init)
    action = aut.action['impl']
    # regression
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "(_hold' = 0) /\ "
        "ite(_hold = 0, x <=> ~ x', x => ~ x')")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)


def test_rabin_persistence():
    aut = trl.default_rabin_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['sys'] = ['x']
    aut.init['env'] = aut.add_expr(' ~ x')
    aut.win['<>[]'] = [aut.add_expr('x')]
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.true, aut.bdd.to_expr(win_set)
    # tranducer
    gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut)
    init = aut.init['impl_env']
    s = '(_goal = 0) /\ (_hold = 1) /\ ~ x'
    init_ = aut.add_expr(s)
    assert init == init_, aut.bdd.to_expr(init)
    action = aut.action['impl']
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "ite(_hold' = 1, ~ x /\ x', x /\ x')")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)
    #
    # unrealizable
    aut.win['[]<>'] = [aut.add_expr(' ~ x')]
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.false, aut.bdd.to_expr(win_set)


def test_rabin_persistence_2():
    aut = trl.default_rabin_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['sys'] = ['x']
    aut.init['sys'] = aut.true
    aut.win['<>[]'] = [aut.add_expr('x'), aut.add_expr(' ~ x')]
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.true, aut.bdd.to_expr(win_set)
    # transducer
    gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut)
    init = aut.init['impl_env']
    s = '(_goal = 0) /\ (_hold = 2)'
    init_ = aut.add_expr(s)
    assert init == init_, aut.bdd.to_expr(init)
    action = aut.action['impl']
    # enumeration.dump_relation(action, aut)
    s = (
        "(_goal = 0) /\ (_goal' = 0) /\ "
        "ite(_hold = 1,"
        "(_hold' = 1) /\ ~ x',"
        "ite(_hold = 0,"
        "(_hold' = 0) /\ x',"
        "(_hold = 2) /\ ("
        "( (_hold' = 0) /\ x' ) \/ "
        "( (_hold' = 1) /\ ~ x' )"
        ") ) )")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)


def test_streett_with_safety_assumption():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['env'] = ['x']
    aut.moore = False
    aut.plus_one = False
    aut.action['env'] = aut.add_expr('x')
    aut.action['sys'] = aut.add_expr('x')
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.true, z
    # transducer
    gr1.make_streett_transducer(z, yij, xijk, aut)
    init = aut.init['impl_env']
    assert init == aut.add_expr('_goal = 0'), aut.bdd.to_expr(init)
    action = aut.action['impl']
    action_ = aut.add_expr("x => ((_goal = 0) /\ (_goal' = 0))")
    assert action == action_, aut.bdd.to_expr(action)
    #
    # negate action to make unrealizable
    aut.action['sys'] = aut.add_expr(' ~ x')
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.add_expr(' ~ x'), aut.bdd.to_expr(z)
    # transducer
    with assert_raises(AssertionError):
        gr1.make_streett_transducer(z, yij, xijk, aut)
    # Moore case
    aut.moore = True
    aut.plus_one = False
    aut.action['env'] = aut.add_expr('x')
    aut.action['sys'] = aut.add_expr('x')
    # solve
    z_moore, yij, xijk = gr1.solve_streett_game(aut)
    assert z_moore != z, 'should differ due to plus_one'
    gr1.make_streett_transducer(z_moore, yij, xijk, aut)
    init = aut.init['impl_env']
    assert init == aut.add_expr('_goal = 0'), aut.bdd.to_expr(init)


def test_rabin_with_safety_assumption():
    aut = trl.default_rabin_automaton()
    aut.declare_variables(x='bool')
    aut.varlist['env'] = ['x']
    aut.init['env'] = aut.add_expr('x')
    aut.action['env'] = aut.add_expr("x' ")
    aut.action['sys'] = aut.add_expr("x")
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.add_expr('x'), win_set
    # transducer
    gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut)
    init = aut.init['impl_env']
    s = 'x /\ (_goal = 0) /\ (_hold = 1)'
    init_ = aut.add_expr(s)
    assert init == init_, aut.bdd.to_expr(init)
    action = aut.action['impl']
    s = "((_goal = 0) /\ (_goal' = 0) /\ (_hold' = 0) /\ x)"
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)
    #
    # negate action to make unrealizable
    aut.action['sys'] = aut.add_expr(' ~ x')
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.false, aut.bdd.to_expr(win_set)
    with assert_raises(AssertionError):
        gr1.make_rabin_transducer(zk, yki, xkijr, aut)


def test_streett_with_liveness_assumption():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool', y=(0, 2))
    aut.varlist = dict(env=['x'], sys=['y'])
    aut.init['env'] = aut.add_expr('y < 2')
    aut.action['sys'] = aut.add_expr(
        """
        /\ ( ((y = 0) /\ ~ x) => (y' = 0) )
        /\ ((y = 0) => (y' < 2))
        /\ ((y = 1) => (y' = 0))
        /\ ((y = 2) => FALSE)
        /\ y \in 0..2
        """)
    aut.win['<>[]'] = [aut.add_expr(' ~ x')]
    aut.win['[]<>'] = [aut.add_expr('y = 1')]
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    z_ = aut.add_expr('y < 2')
    e = aut.bdd.to_expr(z)
    e_ = aut.bdd.to_expr(z_)
    assert z == z_, (e, e_)
    # transducer
    gr1.make_streett_transducer(z, yij, xijk, aut)
    assert action_refined(aut)
    init = aut.init['impl_env']
    assert init == aut.add_expr('(y < 2) /\ (_goal = 0)'), aut.bdd.to_expr(init)
    action = aut.action['impl']
    s = (
        "( (y = 0) => ite(x, (y' = 1), (y' = 0)) ) /\ "
        "( (y = 1) => (y' = 0) ) /\ "
        "( (_goal = 0) /\ (_goal' = 0) ) /\ "
        "( (y /= 2) /\ (y /= 3) )")
    action_ = aut.add_expr(s)
    sat = list(aut.bdd.pick_iter(action))
    sys_action = aut.action['sys']
    sys_action = gr1._copy_bdd(sys_action, aut.bdd, aut.bdd)
    u = aut.apply('=>', action, sys_action)
    assert u == aut.bdd.true, u
    assert action == action_, (action, action_, pprint.pprint(sat))
    #
    # test complement
    b = trl.default_rabin_automaton()
    b.acceptance = 'Rabin(1)'
    b.declare_variables(x='bool', y=(0, 2))
    b.varlist = dict(env=['y'], sys=['x'])
    b.action['env'] = gr1._copy_bdd(aut.action['sys'], aut.bdd, b.bdd)
    b.win['<>[]'] = [b.add_expr('y /= 1')]
    b.win['[]<>'] = [b.add_expr('x')]
    zk, yki, xkijr = gr1.solve_rabin_game(b)
    rabin_win_set = zk[-1]
    bdd = b.bdd
    streett_win_set = gr1._copy_bdd(z, aut.bdd, bdd)
    assert rabin_win_set == bdd.apply('not', streett_win_set)
    with assert_raises(AssertionError):
        gr1.make_rabin_transducer(zk, yki, xkijr, b)


def test_streett_2_goals():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool')
    aut.varlist = dict(env=list(), sys=['x'])
    aut.win['[]<>'] = [aut.add_expr('x'), aut.add_expr(' ~ x')]
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.bdd.true, aut.bdd.to_expr(z)
    # transducer
    gr1.make_streett_transducer(z, yij, xijk, aut)
    assert action_refined(aut)
    action = aut.action['impl']
    # print_fol_bdd(action, aut.bdd, aut.vars)
    # enumeration.dump_relation(action, aut)
    s = (
        "((x /\ (_goal = 0)) => (_goal' = 1)) /\ "
        "(( ~ x /\ (_goal = 1)) => (_goal' = 0)) /\ "
        "(( ~ x /\ (_goal = 0)) => (x' /\ (_goal' = 0))) /\ "
        "((x /\ (_goal = 1)) => ( ~ x' /\ (_goal' = 1)))")
    action_ = aut.add_expr(s)
    assert action == action_, aut.bdd.to_expr(action)


def test_rabin_goal():
    aut = trl.default_rabin_automaton()
    aut.declare_variables(x='bool')
    aut.varlist = dict(env=['x'], sys=list())
    aut.win['[]<>'] = [aut.add_expr('x')]
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.false, aut.bdd.to_expr(win_set)


def test_rabin_2_goals():
    aut = trl.default_rabin_automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist = dict(env=['x'], sys=['y'])
    aut.win['[]<>'] = [aut.add_expr('x => y')]
    # solve
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.true, aut.bdd.to_expr(win_set)
    #
    # unrealizable
    aut.win['<>[]'] = [aut.add_expr(' ~ y')]
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.false, aut.bdd.to_expr(win_set)
    #
    # realizable again
    aut.win['<>[]'] = [aut.add_expr('y')]
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.true, aut.bdd.to_expr(win_set)
    #
    # unrealizable
    aut.win['[]<>'] = [aut.add_expr(s) for s in ['x => y', ' ~ y']]
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.false, aut.bdd.to_expr(win_set)


def test_is_realizable():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool', y=(0, 5))
    aut.varlist = dict(env=['x'], sys=['y'])
    # \A \A realizable
    aut.init['env'] = aut.add_expr('x /\ (y = 1)')
    aut.init['sys'] = aut.add_expr('(y < 3) /\ y \in 0..5')
    aut.qinit = '\A \A'
    aut.build()
    z = aut.add_expr('x /\ (y < 2)')
    assert gr1.is_realizable(z, aut)
    # \A \A unrealizable
    aut.init['sys'] = aut.add_expr('(y < 1)')
    assert not gr1.is_realizable(z, aut)
    # \E \E realizable
    aut.init['env'] = aut.add_expr('x /\ (y = 1)')
    aut.init['sys'] = aut.add_expr('(y < 3)')
    aut.qinit = '\E \E'
    z = aut.add_expr('x /\ (y < 2)')
    assert gr1.is_realizable(z, aut)
    # \E \E unrealizable
    aut.init['env'] = aut.add_expr('x /\ (y = 1)')
    aut.init['sys'] = aut.add_expr('(y > 10)')
    aut.qinit = '\E \E'
    z = aut.true
    assert not gr1.is_realizable(z, aut)
    # \A \E realizable
    aut.moore = False
    aut.init['env'] = aut.true
    s = (
        '(x => (y = 1)) /\ '
        '((~ x) => (y = 4))')
    aut.init['sys'] = aut.add_expr(s)
    aut.qinit = '\A \E'
    z = aut.add_expr('y > 0')
    assert gr1.is_realizable(z, aut)
    # \A \E unrealizable
    s = (
        '(x => (y = 1)) /\ '
        '((~ x) => (y = 10))')
    aut.init['sys'] = aut.add_expr(s)
    aut.qinit = '\A \E'
    z = aut.true
    assert not gr1.is_realizable(z, aut)
    # \E \A realizable
    aut.moore = True
    s = 'x => (y = 1)'
    aut.init['sys'] = aut.add_expr(s)
    aut.qinit = '\E \A'
    z = aut.add_expr('y <= 2')
    assert gr1.is_realizable(z, aut)
    # \E \A unrealizable
    s = (
        '(x => (y = 1)) /\ '
        '((~ x) => (y = 3))')
    aut.init['sys'] = aut.add_expr(s)
    aut.qinit = '\E \A'
    z = aut.true
    assert not gr1.is_realizable(z, aut)


def test_make_init():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool')
    aut.varlist = dict(sys=['x'])
    internal_init = aut.true
    win = aut.true
    aut.qinit = '???'
    with assert_raises(ValueError):
        gr1._make_init(internal_init, win, aut)


def test_streett_qinit_exist_forall():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool', y=(0, 5))
    aut.varlist = dict(env=['x'], sys=['y'])
    aut.moore = True
    aut.qinit = '\E \A'
    aut.init['env'] = aut.add_expr('x')
    aut.init['sys'] = aut.add_expr('(x => (y = 4)) /\ y \in 0..5')
    aut.action['sys'] = aut.add_expr('y \in 0..5')
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.add_expr('y \in 0..5'), aut.bdd.to_expr(z)
    # transducer
    gr1.make_streett_transducer(z, yij, xijk, aut)
    u = aut.init['impl_env']
    u_ = aut.add_expr('x /\ (y = 4) /\ (_goal = 0)')
    assert u == u_, aut.bdd.to_expr(u)


def test_trivial_winning_set():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool')
    aut.varlist = dict(env=list(), sys=['x'])
    aut.win['<>[]'] = [aut.add_expr(' ~ x')]
    triv, aut = gr1.trivial_winning_set(aut)
    assert triv == aut.true, aut.bdd.to_expr(triv)


def test_warn_moore_mealy():
    aut = trl.default_streett_automaton()
    aut.declare_variables(x='bool', y=(0, 25))
    aut.varlist = dict(env=['x'], sys=['y'])
    # Moore OK
    aut.moore = True
    aut.action['env'] = aut.add_expr("x /\ x' ")
    aut.action['sys'] = aut.add_expr("(y > 4) /\ (y' = 5)")
    aut.build()
    r = gr1._warn_moore_mealy(aut)
    assert r is True, r
    # Moore with env' in sys action
    aut.action['env'] = aut.add_expr("x /\ x' ")
    aut.action['sys'] = aut.add_expr("x' /\ (y > 4) /\ (y' = 5)")
    r = gr1._warn_moore_mealy(aut)
    assert r is False, r
    # Mealy with env' in sys action
    aut.moore = False
    r = gr1._warn_moore_mealy(aut)
    assert r is True, r
    # Mealy with sys' in env action
    aut.action['env'] = aut.add_expr("x /\ x' /\ (y' > 4)")
    aut.moore = False
    r = gr1._warn_moore_mealy(aut)
    assert r is False, r
    # Moore with sys' in env action
    aut.action['sys'] = aut.add_expr("(y > 4) /\ (y' = 5)")
    aut.moore = True
    r = gr1._warn_moore_mealy(aut)
    assert r is False, r


def action_refined(aut):
    """Return `True` if action of `'impl'` refines `'sys'`.

    @type aut_a, aut_b: `temporal.Automaton`
    """
    a = aut.action['sys']
    b = aut.action['impl']
    refined = aut.apply('=>', b, a)
    return refined == aut.true, aut.bdd.to_expr(b & ~ a)


def print_fol_bdd(u, bdd, table):
    """Print a first-order formula for node `u`.

    Example:

    ```
    aut = temporal.Automaton()
    aut.vars = dict(x=dict(type='int', dom=(0, 5), owner='sys'))
    aut.action['sys'] = ['x = 1']
    u = aut.action['sys'][0]
    print_fol_bdd(u, aut.bdd, aut.vars)
    ```
    """
    bdd.incref(u)
    dvars = symbolic._prime_and_order_table(table)
    m, umap = mdd.bdd_to_mdd(bdd, dvars)
    u_ = -umap[abs(u)]
    s = m.to_expr(u_)
    print(s)


if __name__ == '__main__':
    test_rabin_2_goals()



"""First-order wrapper for BDDs.

References
==========

Leslie Lamport, Lawrence C. Paulson
    "Should your specification language be typed?"
    ACM Transactions on Programming Languages and Systems
    Vol.21, No.3, pp.502--526, 1999

Kenneth Kunen
    "The foundations of mathematics"
    College Publications, 2009
"""
# Copyright 2016 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
# other places where relevant functions exist:
#   dd.mdd
#   omega.logic.bitvector
#   omega.symbolic.enumeration
#
# examples/interleaving
# symbolic_transducers
# simulate
import logging
import pprint

try:
    from dd import cudd as _bdd
except ImportError:
    from dd import autoref as _bdd

from omega.logic import bitvector as bv
from omega.logic import lexyacc
from omega.logic import syntax as stx
from omega.symbolic import bdd as sym_bdd
from omega.symbolic import cover as cov
from omega.symbolic import enumeration as enum


log = logging.getLogger(__name__)
_parser = lexyacc.Parser()
TYPE_HINTS = {'int', 'saturating', 'modwrap'}


class Context(object):
    """First-order interface to a binary decision diagram.

    All operations assume that integer-valued variables
    take only values that correspond to the Boolean-valued
    variables that refine them.

    Quantification is implicitly bounded.
    In the future, the bound will be made explicit
    in the syntax.

    "Context" alludes to extension of a formal theory
    by definitions [Kunen, Sec.II.15].
    """

    def __init__(self):
        """Instantiate first-order context."""
        self.vars = dict()
        self.bdd = _bdd.BDD()
        self.op = dict()  # operator name -> `str`
        self.op_bdd = dict()  # operator name -> bdd

    def __str__(self):
        return ((
            'Refinement of variables by '
            'Boolean-valued variables:\n\n'
            '{vars}').format(
                vars=pprint.pformat(self.vars)))

    def declare(self, **vrs):
        """Declare variable identifiers given as keyword args.

        Example:

        ```python
        c.declare(x=(2, 15), y='bool', z=(-3, 4))
        ```

        Wrapper of `self.add_vars` that may replace it.
        """
        d = bv.make_symbol_table(vrs)
        self.add_vars(d)

    def add_vars(self, dvars):
        r"""Refine variables in `dvars`.

        The variables in `dvars` should have type hints.
        A Boolean-valued variable remains so.
        An integer-valued variable is assumed to take the
        value resulting as a function of some (fresh)
        Boolean-valued variables.

        Sometimes these variables are called "bits".
        The function is based on two's complement,
        see `omega.logic.bitvector` for details.

        In other words, type hints are used to pick
        a refinement of integers by finitely many bits.
        A sufficient number of bits is selected,
        and operations assume this as type invariant,
        *not* the exact type hint given.

        For example, an integer `x` with type hint `x \in 0..2`
        will be refined using 2 Boolean-valued variables
        `x_0` and `x_1`. All operations and quantification
        will assume that `x \in 0..3`.
        Mind the extra value (3) allowed,
        compared to the hint (0..2).

        Attention:

        - Fine-grained type predicates
          (`n..m` with `n` and `m` other than powers of 2)
          are not managed here.

        - Priming is not reasoned about here.
          Priming is cared for by other modules.

        The method `add_vars` adds to `vars[var]` the keys:

          - `"bitnames"`: `list`
          - `"signed"`: `True` if signed integer
          - `"width"`: `len(bitnames)`
        """
        assert dvars, dvars
        self._avoid_redeclaration(dvars)
        vrs = {k: v for k, v in dvars.items()
               if k not in self.vars}
        if not vrs:
            return
        t = bv.bitblast_table(vrs)
        self.vars.update(t)
        bits = bv.bit_table(t, t)
        for bit in bits:
            self.bdd.add_var(bit)

    def _avoid_redeclaration(self, dvars):
        # if any `dvars` not fresh, then must be same
        gen = (var for var in dvars if var in self.vars)
        for var in gen:
            old = self.vars[var]
            for k, v in dvars[var].items():
                if k in old and v == old[k]:
                    continue
                raise ValueError((
                    'attempted to redeclare "{var}" '
                    'where old: {old} and new: {new}').format(
                        var=var, old=old, new=dvars[var]))

    def support(self, u):
        """Return FOL variables that `u` depends on."""
        supp = self.bdd.support(u)
        bit2int = bv.map_bits_to_integers(self.vars)
        return set(map(bit2int.__getitem__, supp))

    def let(self, defs, u):
        """Return substitution using `defs` in `u`.

        @param defs: `dict` that maps variable names to BDD operators
            or values.
        @param u: BDD operator
        @return: BDD operator
        """
        return self.replace(u, defs)

    def replace(self, u, vars_to_new):
        """Return substitution of var names by values or vars.

        @param vars_to_new: `dict` that maps each var name to
            a var (as `str`), or to a value (as `bool` or `int`).
        """
        if not vars_to_new:
            return u
        assert vars_to_new, vars_to_new
        for k in vars_to_new:
            rename = stx.isinstance_str(vars_to_new[k])
            break
        if rename:
            d = _refine_renaming(vars_to_new, self.vars)
        else:
            d = _refine_assignment(vars_to_new, self.vars)
        return self.bdd.let(d, u)

    def replace_with_bdd(self, u, var_subs):
        """Substitute Boolean-valued variables with BDD nodes."""
        # distinct from `replace` due to restriction to Boolean
        return self.bdd.compose(u, var_subs)

    def forall(self, qvars, u):
        """Universally quantify `qvars` in `u`."""
        r = self.apply('not', u)
        r = self.exist(qvars, r)
        r = self.apply('not', r)
        return r

    def exist(self, qvars, u):
        """Existentially quantify `qvars` in `u`."""
        if not qvars:
            return u
        qbits = bv.bit_table(qvars, self.vars)
        return self.bdd.exist(qbits, u)

    def count(self, u, care_vars=None):
        """Return number of satisfying assignments.

        @param care_vars: variables that the assignments
            should contain. Should be `support(u) <= care_vars`
            If `care_vars == None`, then use `support(u)`.
        """
        # We could allow `support(u) > care_vars`.
        # But that needs a dedicated BDD traversal
        # (to avoid enumeration). Deferred until needed.
        support = self.support(u)
        if care_vars is None:
            care_vars = support
        assert set(care_vars) >= support, (care_vars, support)
        bits = _refine_vars(care_vars, self.vars)
        n = len(bits)
        c = self.bdd.count(u, n)
        assert c == int(c), c
        assert c >= 0, c
        return c

    def pick(self, u, care_vars=None):
        """Return a satisfying assignment, or `None`."""
        return next(self.pick_iter(u, care_vars), None)

    def pick_iter(self, u, care_vars=None):
        """Generator of first-order satisfying assignments."""
        if care_vars is None:
            care_bits = None
        elif care_vars:
            care_bits = bv.bit_table(care_vars, self.vars)
        else:
            care_bits = set()
        for bit_assignment in self.bdd.pick_iter(
                u, care_vars=care_bits):
            for d in enum._bitfields_to_int_iter(
                    bit_assignment, self.vars):
                yield d

    def define(self, e):
        """Register operator definitions.

        The string `e` must contain definitions. Example:

        ```python
        e = '''
            a == x + y > 3
            b == z - x <= 0
            c == a /\ b
            '''
        ```

        In the future, this method may merge with `add_expr`.
        """
        assert stx.isinstance_str(e), e
        bv_defs = bv._parser.parse(e)
        defs = _parser.parse(e)
        for opdef, bv_opdef in zip(defs, bv_defs):
            assert opdef.operator == '==', opdef
            name_ast, expr_ast = opdef.operands
            _, bv_ast = bv_opdef.operands
            name = name_ast.value
            if name in self.vars:
                raise ValueError((
                    'Attempted to define operator "{name}", '
                    'but "{name}" already declared as variable: '
                    '{old}').format(
                        name=name, old=self.vars[name]))
            if name in self.op:
                raise ValueError((
                    'Attempted to redefine operator "{name}". '
                    'Previous definition as: "{old}"').format(
                        name=name, old=self.op[name]))
            s = bv_ast.flatten(
                t=self.vars, defs=self.op_bdd)
            assert stx.isinstance_str(s), s
            # sensitive point:
            #     operator expressions are stored before substitutions
            #     operator BDDs are stored after operator substitutions
            # operator definitions cannot change, so this should
            # not cause problems as currently arranged.
            self.op[name] = expr_ast.flatten()
            self.op_bdd[name] = sym_bdd.add_expr(s, self.bdd)

    def to_bdd(self, expr):
        """Return BDD for the formula `expr`."""
        return self.add_expr(expr)

    def bdds_from(self, *expressions):
        """Return `list` of BDDs for the `expressions`."""
        return [self.to_bdd(e) for e in expressions]

    def add_expr(self, e, with_ops=False):
        """Add first-order predicate.

        A predicate is a Boolean-valued formula.
        """
        assert stx.isinstance_str(e), e
        # optional because current implementation is slow
        if with_ops:
            defs = self.op
        else:
            defs = None
        s = bv.bitblast(e, vrs=self.vars, defs=defs)
        assert stx.isinstance_str(s), s  # was `e` a predicate ?
        return sym_bdd.add_expr(s, self.bdd)

    def to_expr(self, u, care=None, **kw):
        """Return minimal DNF of integer inequalities.

        For now, this method supports all variables in
        `support(u)` being integers.

        @param care: BDD of care set
        @param kw: keyword args are passed to
            function `cover.dumps_cover`.
        """
        if care is None:
            care = self.bdd.true
        cover = cov.minimize(u, care, self)
        s = cov.dumps_cover(
            cover, u, care, self, **kw)
        # prepare to assert
        _, px, _, _ = cov._setup_aux_vars(u, care, self)
        r = cov._list_orthotope_expr(
            cover, px, self, simple=True)
        r = stx.disj(r)
        u_ = self.add_expr(r)
        # promise to match `u` only inside `care`
        # forgive `|#  u => care`
        assert u & care == u_ & care, s
        return s

    def assign_from(self, assignment):
        """Return bdd from `assignment` to variables.

        @param assignment: `dict` that maps each variable to
            a value (`int` or `bool`) in accord with `self.vars`.
        """
        bit_values = _refine_assignment(assignment, self.vars)
        return self.bdd.cube(bit_values)

    def apply(self, op, u, v=None, w=None):
        """Apply operator `op` on operands `u, v, w`."""
        return self.bdd.apply(op, u, v, w)

    @property
    def false(self):
        """Return `self.bdd.false`."""
        return self.bdd.false

    @property
    def true(self):
        """Return `self.bdd.true`."""
        return self.bdd.true


def reorder(dvars, fol):
    """Shift integers up in the variable order of `fol.bdd`."""
    assert dvars, dvars
    bdd = fol.bdd
    for var, d in dvars.items():
        level = d['level']
        dv = fol.vars[var]
        var_type = dv['type']
        if var_type == 'bool':
            bitnames = [var]
        else:
            assert var_type in TYPE_HINTS, var_type
            bitnames = dv['bitnames']
        # print('Shifting bits {bitnames} to '
        #       'level {level}'.format(
        #         level=level,
        #         bitnames=bitnames))
        # assert each bit goes up in level
        for bit in bitnames:
            old_level = bdd.level_of_var(bit)
            assert old_level >= level, (old_level, level)
        # shift to levels wanted
        order = [bdd.var_at_level(i)
                 for i in range(len(bdd.vars))]
        below = order[:level]
        above = order[level:]
        above = [v for v in above if v not in bitnames]
        new_order = below + bitnames + above
        dorder = {var: i for i, var in enumerate(new_order)}
        _bdd.reorder(bdd, dorder)
        order = [bdd.var_at_level(i)
                 for i in range(len(bdd.vars))]
        assert order == new_order, (order, new_order)


def _refine_vars(fol_vars, table):
    """Return bits that represent the `fol_vars`."""
    if fol_vars:
        bits = bv.bit_table(fol_vars, table)
    else:
        # `bit_table` raises `AssertionError` for
        # empty `fol_vars`
        bits = set()
    return bits


def _refine_assignment(m, table):
    """Return bit assignment, from int/bool assignment `m`."""
    bit_values = dict()
    for var, value in m.items():
        assert var in table, var
        if table[var]['type'] == 'bool':
            bit_values[var] = value
            continue
        d = _int_to_bit_assignment(var, value, table)
        bit_values.update(d)
    return bit_values


def _int_to_bit_assignment(var, value, table):
    """Return assignment to bits from assignment to integer.

    Skips the parser, compared to `_assignment_to_bdd`.
    """
    assert var in table, var
    var_bits = bv.var_to_twos_complement(var, table)
    int_bits = bv.int_to_twos_complement(value)
    p, q = bv.equalize_width(var_bits, int_bits)
    values = dict()
    for u, v in zip(p, q):
        if u.isdigit():
            assert u == v, (u, v)
        else:
            values[u] = bool(int(v))
    return values


def _assignment_to_bdd(dvars, fol):
    """Return BDD from assignment to `dvars`.

    Handles only assignments of integer values.
    """
    raise DeprecationWarning('use `_refine_assignment` instead')
    conj = stx.conj(
        '{var} = {value}'.format(var=var, value=value)
        for var, value in dvars.items())
    u = fol.add_expr(conj)
    return u


def _refine_renaming(fol_rename, table):
    """Return renaming of bits, from renaming of FOL vars."""
    bit_rename = dict()
    for old, new in fol_rename.items():
        old_d = table[old]
        new_d = table[new]
        old_type = old_d['type']
        new_type = new_d['type']
        assert old_type == new_type, (old_type, new_type)
        if old_type == 'bool':
            bit_rename[old] = new
            continue
        assert old_type in TYPE_HINTS, old_type
        # int
        old_dom = old_d['dom']
        new_dom = new_d['dom']
        assert old_dom == new_dom, (old_dom, new_dom)
        old_bits = old_d['bitnames']
        new_bits = new_d['bitnames']
        assert len(old_bits) == len(new_bits), (
            old_bits, new_bits)
        # no overlap (otherwise use `compose`)
        common = set(old_bits).intersection(new_bits)
        assert not common, (
            common, old_bits, new_bits)
        # bit substitution
        # assume same bit numbering (index 0 is LSB)
        bit_rename.update(zip(old_bits, new_bits))
    return bit_rename


class Operator(_bdd.Function):
    """Predicate over first-order variables.

    Represented using a binary-decision diagram.
    Pass the relevant `Context` upon instantiation:

    ```python
    c = Context()
    v = Operator(u, c)
    ```
    """

    def __init__(self, *arg, **kw):
        self.fol = kw.pop('fol')
        super(Operator, self).__init__(*arg, **kw)

    @property
    def support(self):
        return self.fol.support(self.node)


def _prime_bits_of_integers(ints, t):
    """Return bit priming for integers in `x`."""
    bit_rename = dict()
    for i in ints:
        bits = t.vars[i]['bitnames']
        d = {k: stx.prime(k) for k in bits}
        bit_rename.update(d)
    return bit_rename


def closed_interval(var, a, b):
    """Return string `a <= var <= b`."""
    return (
        '({a} <= {var}) & '
        '({var} <= {b})').format(
            var=var, a=a, b=b)


def add_one_hot(var, a, b):
    """Return symbol table for one-hot encoding."""
    t = dict()
    for i in range(a, b):
        var = '{var}{i}'.format(var=var, i=i)
        d = dict(type='bool', owner='parameters', level=0)
        t[var] = d
    return t


def array_to_logic(a, counter, aut):
    """Return logic formula for Boolean array `a`.

    No array bounds produced.
    Instead, if `(counter < 0) or (counter >= len(a) - 1)`,
    then select `a[-1]`.
    Therefore, array bounds *must* be added separately.

    @param a: `list` of elements
    @param counter: name of index variable
    """
    bdd = aut.bdd
    r = a[-1]
    for i, x in enumerate(a[:-1]):
        s = '{counter} = {i}'.format(counter=counter, i=i)
        u = aut.add_expr(s)
        r = bdd.ite(u, x, r)
    return r


def multiplexer(a, bits, aut):
    """Return BDD node for selection of elements from `a`.

    The resulting expression is:

    ```
    ite(bits[-1], a[-1],
        ite(bits[-2], a[-2],
            ...
            ite(bits[1], a[1],
                ite(bits[0], a[0], FALSE)
    ```

    This shows that if i < j,
    then b[i] has lower priority than b[j].
    So, this is not exactly a multiplexer.
    """
    assert len(a) == len(bits), (a, bits)
    bdd = aut.bdd
    r = bdd.false
    for bit, x in zip(bits, a):
        g = aut.add_expr(bit)
        r = bdd.ite(g, x, r)
    return r



"""Test the module `omega.steps`."""
from nose import tools as nt
from omega import steps
from omega.symbolic import temporal as trl


def test_step():
    aut = trl.Automaton()
    aut.declare_variables(x='bool', y=(1, 3))
    aut.varlist = dict(env=['x'], sys=['y'])
    aut.prime_varlists()
    aut.init['impl_sys'] = 'True'
    action = aut.add_expr("x /\ (~ x') /\ (y = 2) /\ (y' = 3)")
    aut.action['impl'] = action
    stepper = steps.AutomatonStepper(aut)
    # `action` enabled at `state`
    state = dict(x=True, y=2)
    next_values = stepper.step(state)
    d = dict(x=False, y=3)
    assert next_values == d, (next_values, d)
    # `action` not enabled at `state`
    state = dict(x=True, y=1)
    with nt.assert_raises(ValueError):
        stepper.step(state)


def test_omit_prefix():
    d = {'a': 1, 'foo_mem': 3}
    prefix = 'foo'
    r = steps.omit_prefix(d, prefix)
    r_ = {'a': 1, '_mem': 3}
    assert r == r_, (r, r_)
    # name conflict
    d = {'a': 1, '_mem': 2, 'foo_mem': 3}
    with nt.assert_raises(AssertionError):
        steps.omit_prefix(d, prefix)


if __name__ == '__main__':
    test_component()



"""Syntactic manipulation of trees."""
# Copyright 2014 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import copy
import os
import networkx as nx


class Tree(nx.MultiDiGraph):
    """Abstract syntax tree as a graph data structure.

    Use this as a scaffold for syntactic manipulation.
    It makes traversals and graph rewriting easier,
    so it is preferred to working directly with the
    recursive AST classes.

    The attribute `self.root` is the tree's root node.
    """

    def __init__(self):
        self.root = None
        super(Tree, self).__init__()

    def __repr__(self):
        return repr(self.root)

    def __str__(self):
        # need to override networkx.DiGraph.__str__
        return ('Abstract syntax tree as graph with edges:\n' +
                str([(str(u), str(v))
                    for u, v, k in self.edges(keys=True)]))

    @property
    def variables(self):
        """Return the set of variables in tree."""
        return {u for u in self if u.type == 'var'}

    @classmethod
    def from_recursive_ast(cls, u):
        tree = cls()
        tree.root = u
        tree._recurse(u)
        return tree

    def _recurse(self, u):
        if hasattr(u, 'value'):
            # necessary this terminal is the root
            self.add_node(u)
        elif hasattr(u, 'operator'):
            for i, v in enumerate(u.operands):
                self.add_edge(u, v, key=i)
                self._recurse(v)
        else:
            raise Exception('unknown node type: {u}'.format(u=u))
        return u

    def to_recursive_ast(self, u=None):
        if u is None:
            u = self.root
        w = copy.copy(u)
        if not self.succ.get(u):
            assert hasattr(u, 'value')
        else:
            w.operands = [self.to_recursive_ast(v)
                          for _, v, _ in sorted(
                              self.edges(u, keys=True),
                              key=lambda x: x[2])]
            assert len(u.operands) == len(w.operands)
        return w

    def add_subtree(self, leaf, tree):
        """Add the `tree` at the node `leaf`.

        AST nodes are not copied.
        """
        assert not self.succ.get(leaf)
        for u, v, k in tree.edges(keys=True):
            self.add_edge(u, v, key=k)
        # replace old leaf with subtree root
        ine = self.in_edges(leaf, keys=True)
        if ine:
            assert len(ine) == 1
            ((parent, _, k), ) = ine
            self.add_edge(parent, tree.root, key=k)
        else:
            self.root = tree.root
        self.remove_node(leaf)

    def to_pydot(self, detailed=False):
        """Create GraphViz dot string from given AST."""
        g = ast_to_labeled_graph(self, detailed)
        return nx.drawing.nx_pydot.to_pydot(g)

    def write(self, filename, detailed=False):
        """Layout AST and save result in PDF file."""
        fname, fext = os.path.splitext(filename)
        fext = fext[1:]  # drop .
        p = self.to_pydot(detailed)
        p.set_ordering('out')
        p.write(filename, format=fext)


def ast_to_labeled_graph(tree, detailed):
    """Convert AST to `nextworkx.DiGraph` for graphics."""
    g = nx.DiGraph()
    for u in tree:
        if hasattr(u, 'operator'):
            label = u.operator
        elif hasattr(u, 'value'):
            label = u.value
        else:
            raise TypeError(
                'AST node must be an operator or terminal, '
                'got instead: {u}'.format(u=u) +
                ', of type: {t}'.format(t=type(u)))
        # show both repr and AST node class in each vertex
        if detailed:
            label += '\n' + str(type(u).__name__)
        g.add_node(id(u), label=label)
    for u, v, k in tree.edges(keys=True):
        g.add_edge(id(u), id(v), label=k)
    return g



import logging

import dd.bdd
from nose import tools as nt
from omega.automata import TransitionSystem
from omega.logic import bitvector as bv
from omega.logic import syntax as stx
from omega.symbolic import bdd as sym_bdd
from omega.symbolic import bdd_iterative as bdd_trs
from omega.symbolic import fol as _fol
from omega.symbolic import logicizer
from omega.symbolic import symbolic


log = logging.getLogger('astutils')
log.setLevel('ERROR')
log = logging.getLogger('omega')
log.setLevel('ERROR')


def test_symbolic_automaton():
    aut = symbolic.Automaton()
    aut.vars['x'] = dict(
        type='bool',
        owner='env')
    aut.vars['y'] = dict(
        type='int',
        dom=(3, 17),
        owner='sys')
    aut.init['env'].append('x')
    aut.action['env'].append("x => ~ x'")
    aut.init['sys'].append('y = 5')
    aut.action['sys'].append("(y < 6) => (y' > 10)")
    # strategy definition
    assert aut.moore is True, aut.moore
    assert aut.plus_one is True, aut.plus_one
    assert aut.qinit == '\A \A', aut.qinit
    # str
    s = str(aut)
    assert s.startswith('Symbolic automaton:'), s[:10]
    assert s.endswith('> 10)\n'), s[-3:]
    # data persistence
    s = aut.dumps()
    d = eval(s)
    assert isinstance(d, dict), d
    keys = set(d)
    keys_ = {
        'vars', 'init', 'action', 'win',
        'acceptance', 'moore', 'plus_one', 'qinit'}
    assert keys == keys_, keys


def test_partition_vars():
    bits = ['a', 'b', 'c']
    ubits = ['a']
    ebits = ['b', 'c']
    order = symbolic._add_primed_bits(bits)
    order_ = ["a", "a'", "b", "b'", "c", "c'"]
    assert order == order_, (order, order_)
    prime, partition = symbolic._partition_vars(bits, ubits, ebits)
    prime_ = {k: k + "'" for k in bits}
    partition_ = dict(
        uvars={"a"},
        upvars={"a'"},
        evars={"b", "c"},
        epvars={"b'", "c'"})
    assert prime == prime_, (prime, prime_)
    assert partition == partition_, (partition, partition_)


def test_slugsin_parser():
    parser = sym_bdd.Parser()
    s = '! x'
    u = parser.parse(s)
    assert_is_operator(u, '!')
    (x,) = u.operands
    assert_is_var(x, 'x')
    s = '& x y'
    u = parser.parse(s)
    assert_is_operator(u, '&')
    x, y = u.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '| x y'
    u = parser.parse(s)
    assert_is_operator(u, '|')
    x, y = u.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '^ x12 y0'
    u = parser.parse(s)
    assert_is_operator(u, '^')
    x12, y0 = u.operands
    assert_is_var(x12, 'x12')
    assert_is_var(y0, 'y0')
    s = '$1 0'
    u = parser.parse(s)
    assert_is_buffer(u, 1)
    (a,) = u.memory
    assert_is_number(a, 0)
    s = '$2 0 1'
    u = parser.parse(s)
    assert_is_buffer(u, 2)
    a, b = u.memory
    assert_is_number(a, 0)
    assert_is_number(b, 1)
    s = '$2 $2 1 ?0 ?0'
    u = parser.parse(s)
    assert_is_buffer(u, 2)
    a, b = u.memory
    assert_is_register(b, 0)
    assert_is_buffer(a, 2)
    a, b = a.memory
    assert_is_number(a, 1)
    assert_is_register(b, 0)
    s = '\E x & x y'
    u = parser.parse(s)
    assert_is_operator(u, '\E')
    qvars, e = u.operands
    assert_is_var(qvars, 'x')
    assert_is_operator(e, '&')
    x, y = e.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '\A y | y ! x4'
    u = parser.parse(s)
    assert_is_operator(u, '\A')
    qvars, e = u.operands
    assert_is_var(qvars, 'y')
    assert_is_operator(e, '|')
    y, not_x4 = e.operands
    assert_is_var(y, 'y')
    assert_is_operator(not_x4, '!')
    (x4,) = not_x4.operands
    assert_is_var(x4, 'x4')
    s = '\E x \A y | x y'
    u = parser.parse(s)
    assert_is_operator(u, '\E')
    qvars, e = u.operands
    assert_is_var(qvars, 'x')
    assert_is_operator(e, '\A')
    qvars, e = e.operands
    assert_is_var(qvars, 'y')
    assert_is_operator(e, '|')
    x, y = e.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '\S $4 x y  z w  & y w'
    u = parser.parse(s)
    assert_is_operator(u, '\S')
    subs, e = u.operands
    assert_is_buffer(subs, 4)
    x, y, z, w = subs.memory
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    assert_is_var(z, 'z')
    assert_is_var(w, 'w')
    assert_is_operator(e, '&')
    y, w = e.operands
    assert_is_var(y, 'y')
    assert_is_var(w, 'w')


def assert_is_operator(u, operator):
    assert hasattr(u, 'type'), u
    assert u.type == 'operator', u.type
    assert u.operator == operator


def assert_is_var(u, var):
    assert hasattr(u, 'type'), u
    assert u.type == 'var', u.type
    assert u.value == var, u.value


def assert_is_number(u, j):
    assert hasattr(u, 'type'), u
    assert u.type == 'num', u.type
    assert u.value == str(j), u.value


def assert_is_buffer(u, n):
    assert hasattr(u, 'type'), u
    assert u.type == 'buffer', u.type
    assert hasattr(u, 'memory'), u
    assert len(u.memory) == n, u.memory


def assert_is_register(u, idx):
    assert hasattr(u, 'type'), u
    assert u.type == 'register', u.type
    assert u.value == str(idx), u.value


def test_iterative_bddizer():
    add = bdd_trs.add_expr
    order = {'x': 0, 'y': 1, 'z': 2}
    # x & y
    bdd = dd.bdd.BDD(order)
    e = '& x y'
    u = add(e, bdd)
    v = bdd.add_expr('x & y')
    assert u == v, (u, v)
    # buffers
    # (x & y) | ! z
    bdd = dd.bdd.BDD(order)
    e = '$ 3   & x y   ! z  | ? 0 ? 1'
    u = add(e, bdd)
    v = bdd.add_expr('(x /\ y) \/ ~ z')
    assert u == v, (u, v)


def test_bddizer_propositional():
    add = sym_bdd.add_expr
    order = {'x': 0, 'y': 1, 'z': 2}
    # x & y
    bdd = dd.bdd.BDD(order)
    e = '& x y'
    u = add(e, bdd)
    v = bdd.add_expr('x /\ y')
    assert u == v, (u, v)
    # buffers
    # (x & y) | ! z
    bdd = dd.bdd.BDD(order)
    e = '$ 3   & x y   ! z  | ? 0 ? 1'
    u = add(e, bdd)
    v = bdd.add_expr('(x & y) | ! z')
    assert u == v, (u, v)
    #
    e = '& $2 & 1 x ?0 $3 | !x y z & ?1 z'
    s = 'x /\ z'
    u = add(e, bdd)
    v = bdd.add_expr(s)
    assert u == v, (u, v)


def test_bddizer_quantifiers():
    add = sym_bdd.add_expr
    order = {'x': 0, 'y': 1, 'z': 2}
    bdd = dd.bdd.BDD(order)
    e = '\E x 1'
    u = add(e, bdd)
    assert u == bdd.true, u
    e = '\E x 0'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\A x 1'
    u = add(e, bdd)
    assert u == bdd.true, u
    e = '\A x 0'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\A x x'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\E x x'
    u = add(e, bdd)
    assert u == bdd.true, u
    e = '\A & x y y'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\A x y'
    u = add(e, bdd)
    u_ = bdd.var('y')
    assert u == u_, (u, u_)
    e = '\E & x y & y x'
    u = add(e, bdd)
    assert u == bdd.true, u


def test_bddizer_substitution():
    add = sym_bdd.add_expr
    order = {'x': 0, 'y': 1, 'z': 2, 'w': 3}
    bdd = dd.bdd.BDD(order)
    e = '\S $2 x y y'
    u = add(e, bdd)
    u_ = bdd.var('x')
    assert u == u_, (u, u_)
    e = '\S $4 x y  z w  | y ! w'
    u = add(e, bdd)
    u_ = bdd.add_expr('x | ! z')
    assert u == u_, (u, u_)
    e = '$2 0 \S $2 x y y'
    u = add(e, bdd)
    u_ = bdd.var('x')
    assert u == u_, (u, u_)


def slugsin_parser(s, t):
    """Helper that converts prefix to infix syntax for readability."""
    slugs_table = t.to_slugs()
    ltl_parser = bv.Parser()
    p = ltl_parser.parse(s)
    bv.add_bitnames(slugs_table)
    s = p.flatten(t=slugs_table)
    slugsin_parser = sym_bdd.Parser()
    print(slugsin_parser.parse(s))


def test_logicizer_env():
    g = TransitionSystem()
    g.vars['x'] = 'bool'
    g.owner = 'env'
    g.add_edge(0, 1, x=True)
    g.add_edge(1, 2, formula="x'")
    g.add_edge(2, 1)
    (env_init, env_act,
     sys_init, sys_act) = logicizer._graph_to_formulas(
        g,
        nodevar='k',
        ignore_initial=True,
        self_loops=True)
    s = stx.conj(env_act)
    e1 = "(((((k = 0)) => (((x <=> True)) /\ ((k' = 1)))) \n"
    e2 = "(((((k = 0)) => (((k' = 1)) /\ ((x <=> True)))) \n"
    assert s.startswith(e1) or s.startswith(e2), s
    e3 = (
        "/\ (((k = 1)) => ((x') /\ ((k' = 2))))) \n"
        "/\ (((k = 2)) => ((k' = 1)))) \/ (k' = k)")
    assert s.endswith(e3), s
    # test automaton
    aut = logicizer.graph_to_logic(
        g, nodevar='k', ignore_initial=True)
    assert 'x' in aut.vars, aut.vars
    assert 'k' in aut.vars, aut.vars
    xtype = aut.vars['x']['type']
    ktype = aut.vars['k']['type']
    assert xtype == 'bool', xtype
    assert ktype == 'int', ktype


def test_joint_support():
    fol = setup()
    r = sym_bdd.joint_support([fol.bdd.true], fol)
    r_ = set()
    assert r == r_, (r, r_)
    u = fol.add_expr('x <= 3')
    r = sym_bdd.joint_support([u], fol)
    r_ = {'x'}
    assert r == r_, (r, r_)
    v = fol.add_expr('~ y')
    r = sym_bdd.joint_support([v], fol)
    r_ = {'y'}
    assert r == r_, (r, r_)
    r = sym_bdd.joint_support([u, v], fol)
    r_ = {'x', 'y'}
    assert r == r_, (r, r_)
    u |= v
    r = sym_bdd.joint_support([u], fol)
    r_ = {'x', 'y'}
    assert r == r_, (r, r_)


def test_assert_support():
    fol = setup()
    u = fol.add_expr('x = -3 /\ ~ y')
    assert sym_bdd.support_issubset(u, ['x', 'y'], fol)
    assert not sym_bdd.support_issubset(u, ['x'], fol)


def test_is_state_predicate():
    fol = setup()
    s = "x = -3  /\  ~ y"
    u = fol.add_expr(s)
    assert sym_bdd.is_state_predicate(u)
    assert not sym_bdd.is_primed_state_predicate(u)
    assert not sym_bdd.is_proper_action(u)
    s = "x' = -3  /\  ~ y'"
    u = fol.add_expr(s)
    assert not sym_bdd.is_state_predicate(u)
    assert sym_bdd.is_primed_state_predicate(u)
    assert not sym_bdd.is_proper_action(u)
    s = "x = -3  /\  ~ y'"
    u = fol.add_expr(s)
    assert not sym_bdd.is_state_predicate(u)
    assert not sym_bdd.is_primed_state_predicate(u)
    assert sym_bdd.is_proper_action(u)


def test_prime_unprimed():
    fol = setup()
    s = "x = -3  /\  ~ y"
    u = fol.add_expr(s)
    r = sym_bdd.prime(u, fol)
    s = "x' = -3  /\  ~ y'"
    r_ = fol.add_expr(s)
    assert r == r_, fol.bdd.to_expr(r)
    with nt.assert_raises(AssertionError):
        sym_bdd.prime(r, fol)
    with nt.assert_raises(AssertionError):
        sym_bdd.unprime(u, fol)
    s = "x' = -3  /\  ~ y'"
    u = fol.add_expr(s)
    r = sym_bdd.unprime(u, fol)
    s = "x = -3  /\  ~ y"
    r_ = fol.add_expr(s)
    assert r == r_, fol.bdd.to_expr(r)


def setup():
    fol = _fol.Context()
    t = {
        "x": (-4, 5),
        "y": 'bool',
        "x'": (-4, 5),
        "y'": 'bool'}
    fol.declare(**t)
    print(fol.vars)
    return fol



"""Tests for `omega.logic.lexyacc."""
from omega.logic import lexyacc


parser = lexyacc.Parser()


def test_quantifiers():
    s = '\E x:  True'
    t = parser.parse(s)
    assert hasattr(t, 'operator'), t
    assert t.type == 'operator', t
    assert t.operator == '\E', t.operator
    assert len(t.operands) == 2, t.operands
    qvars, e = t.operands
    assert len(qvars) == 1, qvars
    (x,) = qvars
    _assert_is_var_node(x, 'x')
    assert hasattr(e, 'type'), e
    assert e.type == 'bool', e.type
    assert e.value == 'True', e.value
    s = '\E x, y:  False'
    t = parser.parse(s)
    assert hasattr(t, 'type'), t
    assert t.type == 'operator', t.type
    assert t.operator == '\E', t.operator
    assert len(t.operands) == 2, t.operands
    qvars, e = t.operands
    assert len(qvars) == 2, qvars
    x, y = qvars
    _assert_is_var_node(x, 'x')
    _assert_is_var_node(y, 'y')
    assert hasattr(e, 'type'), e
    assert e.type == 'bool', e.type
    assert e.value == 'False', e.value
    s = '\A y:  True'
    t = parser.parse(s)
    assert t.operator == '\A', t.operator
    s = '\A x, y, z:  (x \/ ~ y) /\ z'
    t = parser.parse(s)
    assert t.operator == '\A', t.operator
    assert len(t.operands) == 2, t.operands
    qvars, e = t.operands
    assert len(qvars) == 3, qvars
    r = e.flatten()
    r_ = '( ( x \/ ( ~ y ) ) /\ z )'
    assert r == r_, (r, r_)


def test_substitution():
    s = '\S a / b: True'
    t = parser.parse(s)
    assert hasattr(t, 'type'), t
    assert t.type == 'operator', t.type
    assert t.operator == '\S', t.operator
    assert len(t.operands) == 2, t.operands
    pairs, e = t.operands
    assert len(pairs) == 1, pairs
    (ab,) = pairs
    a, b = ab
    _assert_is_var_node(a, 'a')
    _assert_is_var_node(b, 'b')
    assert hasattr(e, 'type'), e
    assert e.type == 'bool', e.type
    assert e.value == 'True', e.value
    s = '\S a / b,  c / d: False'
    t = parser.parse(s)
    assert hasattr(t, 'type'), t
    assert t.type == 'operator', t
    assert t.operator == '\S', t.operator
    assert len(t.operands) == 2, t.operands
    pairs, e = t.operands
    assert len(pairs) == 2, pairs
    ab, cd = pairs
    assert len(ab) == 2, ab
    a, b = ab
    _assert_is_var_node(a, 'a')
    _assert_is_var_node(b, 'b')
    c, d = cd
    _assert_is_var_node(c, 'c')
    _assert_is_var_node(d, 'd')
    assert hasattr(e, 'type')
    assert e.type == 'bool', e.type
    assert e.value == 'False', e.value


def test_range():
    s = r'z \in 0 .. 2'
    r = parser.parse(s).flatten()
    r_ = r'( z \in ( 0 .. 2 ) )'
    assert r == r_, r
    s = r'''
           z \in 0 .. 2
        /\ y \in -45 .. 1
        '''
    r = parser.parse(s).flatten()
    r_ = (
        '( ( z \in ( 0 .. 2 ) )'
        ' /\ ( y \in ( -45 .. 1 ) ) )')
    assert r == r_, r


def test_multiline_tla_comment():
    # on single line
    s = '(* this is a comment *) a = 1'
    r = parser.parse(s).flatten()
    r_ = '( a = 1 )'
    assert r == r_, r
    # spread over multiple lines
    s = '(* hohoho \n bar bar foo *) a = 1'
    r = parser.parse(s).flatten()
    r_ = '( a = 1 )'
    assert r == r_, r
    # repeated
    s = '(* foo *) a = 1 (* bar *)'
    r = parser.parse(s).flatten()
    r_ = '( a = 1 )'
    assert r == r_, r


def _assert_is_var_node(x, var):
    assert hasattr(x, 'type'), x
    assert x.type == 'var', x.type
    assert x.value == var, x.value


if __name__ == '__main__':
    test_range()



"""Dynamical system representation in first-order logic.

References
==========

Leslie Lamport
    "The temporal logic of actions"
    ACM Transactions on Programming
    Languages and Systems (TOPLAS),
    Vol.16, No.3, pp.872--923, 1994

Amir Pnueli
    "The temporal semantics of concurrent programs"
    Theoretical Computer Science (TCS)
    Vol.13, No.1, pp.45--60, 1981
"""
# Copyright 2015-2017 by California Institute of Technology
# All rights reserved. Licensed under 3-clause BSD.
#
import copy
import logging
import pprint
import weakref

from omega.logic import bitvector as bv
from omega.logic import syntax as stx
from omega.symbolic import bdd as sym_bdd
from omega.symbolic import fol as _fol
from omega.symbolic import prime as prm
from omega.symbolic import symbolic as _sym


log = logging.getLogger(__name__)


class Automaton(_fol.Context):
    """A changing system.

    Both interleaving and noninterleaving style specifications
    can be written using this class.


    User-defined attributes
    =======================

    You define the attributes:

      - `vars`: `dict` that maps each variable name to
        a `dict` with keys:

          - `"type": "boolean" or "modwrap" or "saturating"`
          - `"dom": (min, max)`, range of integer
          - `"init"`: has suitable `__str__` (and is optional)

      - `varlist`: `dict(list)` that maps strings to lists of
        variable names (as strings).
        Usually used to store variables that represent each component.

      - `init`: initial condition
        `ExprDict` that maps strings to BDD nodes.
        Usually the strings are operator names.

      - `action`: transition predicate
        `ExprDict` that maps strings to BDD nodes.
        Usually the strings are operator names.

      - `win`: winning condition. In general, it takes the form:

        ```
        {
            component_name:
                {'<>[]': `list` of BDD nodes,
                 '[]<>': `list` of BDD nodes}
            another_component_name:
                {'<>[]': `list` of BDD nodes,
                 '[]<>': `list` of BDD nodes},
                ...,
            '[]<>': `list` of BDD nodes,
            '<>[]': `list` of BDD nodes
        }
        ```

        Not all these keys need to be present at once.
        Different use cases require different keys.
        For example, a synthesizer accesses the keys
        `'[]<>'` and `'<>[]'`.

      - `bdd`: change this only if you want a different BDD manager.

    The attributes `init` and `action` convert expressions to BDDs,
    for example:

    ```python
    aut.init['env'] = 'x = 0'
    >>> aut.init
    {'env': Function (DdNode) with var index: 3,
            ref count: 2, int repr: 41529443}
    ```

    The attribute `win` is a `dict` and you have to do the conversions.
    Use the method `bdds_from` to simplify code, for example:

    ```python
    aut.win['[]<>'] = aut.bdds_from('x = 0', 'y > 1')
    ```
    """
    # Relation to `omega.symbolic.symbolic.Automaton`
    # ===============================================
    #
    # - uvars = varlists["env"]
    # - upvars = varlists["env'"]
    # - evars = varlists["sys"]
    # - epvars = varlists["sys'"]
    #
    # Memoizing the following is unnecessary:
    #
    #   ubvars, ebvars, uevars, uepvars
    #
    # They are rarely used. Their relevance is to priming and
    # unpriming, which is better done by suitable functions
    # on identifiers.

    # Metalinguistic purposes served by attributes
    # stored in an `Automaton`:
    #
    # - substitution (renaming maps)
    # - grouping variables to feed quantifiers later
    # - defining operators.

    # Note: The keys of the attributes `init`, `action`
    # and `win` can be regarded also as operators.
    # In other words, `init['env']` serves as `EnvInit`,
    # and `action['env']` as `EnvNext`.

    def __init__(self):
        super(Automaton, self).__init__()
        # `varlist` says which variables represent each component
        # essentially, `varlist` is Lamport's `\mu`
        self.varlist = dict()
            # groups of variables for various uses
        self.players = dict()  # player (str) -> turn (int)
        # state machine attributes
        self.init = ExprDict(self)
        self.action = ExprDict(self)
        self.win = dict()
        # BDD cache
        self._bdd_to_expr = dict()  # BDD node -> `str`

    def __copy__(self):
        other = type(self)()
        other.vars = copy.deepcopy(self.vars)
        other.bdd = self.bdd
        other.varlist = copy.deepcopy(self.varlist)
        other.players = copy.deepcopy(self.players)
        other.op = copy.deepcopy(self.op)
        other.op_bdd = copy.copy(self.op_bdd)
        # BDD nodes
        other.init = ExprDict(other, self.init)
        other.action = ExprDict(other, self.action)
        other.win = _copier(self.win)
        return other

    def __str__(self):
        c = list()
        s = 'Variables: \n {v}'.format(
            v=pprint.pformat(self.vars))
        c.append(s)
        for k, v in self.init.items():
            e = self._fetch_expr(v)
            if e is None:
                e = v
            s = 'init[{k}] = {e}'.format(k=k, e=e)
            c.append(s)
        for k, v in self.action.items():
            e = self._fetch_expr(v)
            if e is None:
                e = v
            s = 'action[{k}] = {e}'.format(k=k, e=e)
            c.append(s)
        c.append(str(self.win))
        return 'Automaton:\n' + '\n'.join(c)

    def declare_variables(self, **vrs):
        """Declare unprimed and primed identifiers.

        These identifiers represent the values of
        (flexible) variables.
        """
        d = bv.make_symbol_table(vrs)
        self.add_vars(d, flexible=True)

    def declare_constants(self, **constants):
        """Declare unprimed identifiers.

        These identifiers represent constants,
        also known as rigid variables.
        """
        d = bv.make_symbol_table(constants)
        self.add_vars(d, flexible=False)

    def add_vars(self, vrs, flexible=False):
        """Declare variables `vrs`.

        Refine integer-valued variables by
        Boolean-valued variables, and create maps for
        concrete variables (bits).

        @param flexible: if `True`, then add also variables
            with primed names. Assumes that `vrs` contains
            unprimed names.
        """
        assert vrs, vrs
        assert isinstance(flexible, bool), flexible  # catch kw conflict
        if flexible:
            vrs = _sym.add_primed_too(vrs)
        super(Automaton, self).add_vars(vrs)

    @property
    def vars_of_all_players(self):
        """Set of variables of all players."""
        return self.vars_of_players(self.players)

    def vars_of_players(self, players):
        """Set of variables controlled by `players`."""
        gen = (
            v for k, v in self.varlist.items()
            if k in players)
        return set().union(*gen)

    def prime_varlists(self, keys=None):
        """Map primed `keys` to lists of primed variables.

        For each `k in keys`, add `"{k}'".format(k=k)` to
        `self.varlist`, with value the `list` that results
        from priming the variables in `self.varlist[k]`.

        If `keys is None`, prime all unprimed variable lists.
        """
        if keys is None:
            keys = set(self.varlist)
        for k in keys:
            if stx.isprimed(k):
                continue
            pk = stx.prime(k)
            pvrs = stx.prime_vars(self.varlist[k])
            self.varlist[pk] = pvrs

    def replace_with_primed(self, vrs, u):
        r"""Substitute primed for unprimed `vrs` in `u`.

        For example:

        ```
        u = aut.add_expr("x /\ y /\ z")
        vrs = ['x', 'y']
        v = aut.replace_with_primed(vrs, u)
        v_ = aut.add_expr("x' /\ y' /\ z")
        assert v == v_
        ```

        Identifiers in `vrs` should be declared as variables.
        Identifiers declared as constants will raise errors.
        """
        let = {k: stx.prime(k) for k in vrs}
        return self.let(let, u)

    def replace_with_unprimed(self, vrs, u):
        """Substitute unprimed `vrs` for primed in `u`."""
        let = {stx.prime(k): k for k in vrs}
        return self.let(let, u)

    def implies_type_hints(self, u, vrs):
        """Return `True` if `u => TypeInv` for all vars.

        All declared variables and constants are taken into
        account.
        """
        # not named `assert_type_invariant` because the
        # assertion is about the state predicate `u`,
        # so a static conclusion.
        vrs = {var for var in self.vars
               if not stx.isprimed(var)}
        type_hints = _conjoin_type_hints(vrs, self)
        r = type_hints | ~ u
        return r == self.true

    def type_hint_for(self, vrs):
        """Return initial predicate using type hints for `vrs`.

        See also `self.type_action_for`.
        """
        return self._type_hints_to_formulas(vrs, action=False)

    def type_action_for(self, vrs):
        r"""Return action using type hints for `vrs`.

        The type action has the form:

            TypeAction == Inv /\ Inv'

        where `Inv` is a box. A box is a conjunction of integer
        intervals, one for each variable. For each integer variable,
        the interval constraint has the form:

            var \in min_value .. max_value

        The implementation supports only integer variables,
        so the interval constraint is implemented as the
        conjunction of two inequalities:

            /\ min_value <= var
            /\ var <= max_value

        If we take the closure, the second conjunct will
        result from the type invariant. The above is usable
        with either approach (with or without closure).

        We omit the type action for Boolean variables,
        because we use BDDs underneath.

        @return: formula as `str`
        """
        # to obtain a BDD, pass the result to the method `to_bdd`
        return self._type_hints_to_formulas(vrs, action=True)

    def _type_hints_to_formulas(self, vrs, action):
        r"""Return type constraint for `vrs` as `str`.

        If `action is  True` then return type invariant `Inv`,
        else the action `Inv /\ Inv'`.
        """
        r = list()
        for var in vrs:
            hints = self.vars[var]
            if hints['type'] == 'bool':
                continue
            assert hints['type'] == 'int', hints
            a, b = hints['dom']
            s = r'({a} <= {var}) /\ ({var}  <= {b})'
            type_inv = s.format(a=a, b=b, var=var)
            r.append(type_inv)
            if not action:
                continue
            type_inv_primed = s.format(
                a=a, b=b, var=stx.prime(var))
            r.append(type_inv_primed)
        return stx.conj(r)

    def _cache_expr(self, expr):
        """Overwrite cache entry for BDD described by `expr`."""
        # If the `uid` for this `expr` is absent from the cache,
        # then no formula with this BDD has been added before,
        # or was, then became invalid, and has been garbage collected.
        #
        # If the `uid` is present in the cache,
        # then there are two cases:
        #
        # If the cache entry is valid, then it can only
        # have remained valid since added (or invalidated and
        # re-added by coincidence).
        #
        # However, the same `uid` can result from expressions
        # *other* than `expr`. In that case, the cache is
        # overwritten by a semantically equivalent expression.
        #
        # If the cache entry is invalid, then it should be
        # garbage collected, which leads to the case of
        # the pair `uid, expr` being absent from the cache.
        u = self._add_expr(expr)
        # We store the node ID in order to avoid incrementing the
        # reference count, to allow for garbage collection of the BDD.
        # Thus, caching here leaves the BDD manager behavior
        # unaffected. Otherwise we would produce eternal references.
        uid = self._uid_of_node(u)
        # overwritten ?
        old_expr = self._bdd_to_expr.get(uid)
        if old_expr is not None:
            log.info((
                'BDD cache entry for expression "{old}" '
                'overwritten by expression "{new}"'
                ).format(old=old_expr, new=expr))
        self._bdd_to_expr[uid] = expr
        return u

    def _fetch_expr(self, u):
        """Return cached expression for BDD node `u`.

        Return `None` if no expression remains cached for `u`.
        """
        uid = self._uid_of_node(u)
        expr = self._bdd_to_expr.get(uid)
        # The BDD `u` was never cached (user never entered it
        # via `add_expr`), or has been garbage collected and
        # then the invalid cache cleared.
        if expr is None:
            return None
        # outdated cache ?
        #
        # This can happen if all refs to `u` had been deleted,
        # then `u` was garbage collected, and the ID
        # assigned to a new node.
        #
        # It is possible (though unlikely) that `u` is
        # garbage collected, and the same ID assigned later
        # to the same Boolean formula.
        u_ = self._add_expr(expr)
        # valid cache ?
        if u == u_:
            return expr
        # It is impossible that the user has added `u`,
        # and then the cache became invalid.
        # The cache can become invalid only if the user
        # deletes all their references to `u`.
        #
        # If they do so, then it means that they aren't interested
        # in `u`, and `u` isn't referenced anywhere any more.
        #
        # Of course, `u` could then be computed via some
        # operations, but in that case it is a result,
        # not user input any more.
        #
        # Also, if the user deletes all references, and the
        # cache becomes invalid, and then the user adds the
        # same expression, then the expression will be
        # cached again, in a new valid cache entry.
        #
        # rm invalid cache entry
        self._bdd_to_expr.pop(uid)

    def _clear_invalid_cache(self):
        """Pop any BDD cache entries no longer valid."""
        invalid = set()
        for uid, expr in self._bdd_to_expr.items():
            u = self._add_expr(expr)
            uid_ = self._uid_of_node(u)
            # valid cache entry ?
            if uid == uid_:
                continue
            invalid.add(uid)
        # rm invalid cache entries
        [self._bdd_to_expr.pop(uid) for uid in invalid]

    def _uid_of_node(self, u):
        """Return unique identifier `str` of BDD node `u`."""
        return str(u)

    # lesson: don't call this, because the `uid`
    # may be invalid in the BDD manager.
    def _node_with_uid(self, uid):
        """Return BDD node identified by the string `uid`."""
        return self.bdd.add_expr(uid)

    def _add_expr(self, expr):
        """Return BDD for expression without caching it.

        Use `self.op_bdd` or `self.add_expr`.
        """
        if expr in self.op_bdd:
            return self.op_bdd[expr]
        return self.add_expr(expr)

    def build(self):
        """Prime variable lists in `self.varlist`."""
        self.prime_varlists(['env', 'sys'])


def _assert_invariant(components, aut, moore=True):
    """Assert invariant about `aut.varlists` and predicates.

    - varlists of `components` are disjoint
    - `aut.init` and `aut.win` contain state predicates

    If `moore=True`, then assert that `aut.action` are actions
    of `components`.
    """
    varlists = {k: v for k, v in aut.varlist.items()
                if k in components}
    actions = {k: v for k, v in aut.actions.items()
               if k in components}
    assert prm.pairwise_disjoint(varlists), varlists
    for u in aut.init.values():
        assert sym_bdd.is_state_predicate(u)
    _assert_is_win(aut.win)
    if not moore:
        return
    for component, action in actions.items():
        assert prm.is_action_of_player(action, component, aut)


def _assert_is_win(d):
    if isinstance(d, list):
        for u in d:
            assert sym_bdd.is_state_predicate(u), u
            return
    assert isinstance(d, dict), d
    for _, v in d.items():
        _assert_is_win(v)


class ExprDict(dict):
    """A `dict` that converts expressions to BDDs.

    Setting a string as value triggers a conversion to
    a BDD first, then storage. To reverse this conversion
    call the method `_fetch_expr` of the associated context.

    This class treats values other than `str` the same
    way that `dict` does.
    """

    def __init__(self, context, *arg, **kw):
        super(ExprDict, self).__init__(*arg, **kw)
        self._context = weakref.proxy(context)

    def __setitem__(self, k, v):
        """Set key `k` to value `v`.

        If `v` is a `str`, then convert to BDD,
        then set `k` to the resulting BDD.
        Otherwise behave as a `dict`.
        """
        try:
            v + ''
        except TypeError:
            super(ExprDict, self).__setitem__(k, v)
            return
        if v in self._context.op_bdd:
            u = self._context.op_bdd[v]
        else:
            u = self._context._cache_expr(v)
        super(ExprDict, self).__setitem__(k, u)

    def update(self, *arg, **kw):
        if arg:
            (d,) = arg
            kw.update(d)
        for k, v in kw.items():
            self[k] = v

    def setdefault(self, key, value=None):
        """Return self[key], after setting if needed.

        If `key not in self` and `value` is a `str`,
        then the returned value is a BDD.
        See `__setitem__` for more details.
        """
        if key not in self:
            self[key] = value
        return self[key]


def conj_actions_of(players, aut):
    """Return conjunction of actions from `players`."""
    action = aut.true
    for p in players:
        action &= aut.action[p]
    return action


def _conjoin_type_hints(vrs, fol):
    """Return conjunction of type hints for `vrs` as BDD."""
    r = list()
    for var in vrs:
        hints = fol.vars[var]
        if hints['type'] == 'bool':
            # The constraint `var \in BOOLEAN` will
            # anyway dissapear at the BDD layer.
            continue
        assert hints['type'] == 'int', hints
        a, b = hints['dom']
        s = r'({a} <= {var}) /\ ({var} <= {b})'
        type_hints = s.format(a=a, b=b, var=var)
        r.append(type_hints)
    u = fol.add_expr(stx.conj(r))
    return u


# serves the same purpose as `fill_blanks` in `symbolic`.
def default_streett_automaton():
    """Return an `Automaton` populated as Streett spec."""
    aut = _default_safety_automaton()
    aut.win['[]<>'] = [aut.true]
    aut.win['<>[]'] = [aut.false]
    return aut


def default_rabin_automaton():
    """Return an `Automaton` populated as Rabin spec."""
    aut = _default_safety_automaton()
    aut.win['[]<>'] = [aut.true]
    aut.win['<>[]'] = [aut.true]
    return aut


def _default_safety_automaton():
    """Return a populated `Automaton`, except for liveness."""
    aut = Automaton()
    aut.varlist = dict(env=list(), sys=list())
    aut.init = dict(env=aut.true, sys=aut.true)
    aut.action = dict(env=aut.true, sys=aut.true)
    aut.qinit = r'\E \A'
    aut.moore = True
    aut.plus_one = True
    return aut


def _copier(d):
    """Recursively copy `d`, used for `self.win`."""
    if isinstance(d, list):
        return d.copy()
    assert isinstance(d, dict), d
    r = dict()
    for k, v in d.items():
        r[k] = _copier(v)
    return r



"""Manipulating first-order formulas using binary decision diagrams."""
from omega.symbolic import fol as _fol
from omega.symbolic import temporal as trl


def example_usage_of_ordinary_context():
    ctx = _fol.Context()
    ctx.declare(x=(1, 3), y='bool')


def example_uage_of_modal_context():
    aut = trl.Automaton()
    aut.declare_constants(a=(1, 5), b=(-3, 4), c='bool')
    aut.declare_variables(x=(4, 14), y=(-7, 8), z='bool')


if __name__ == '__main__':
    example_usage_of_ordinary_context()
    example_uage_of_modal_context()



#!/usr/bin/env python
"""Parser for past and future linear temporal logic.

The syntax is based on that of SPIN.
"""
# Copyright 2014 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
# the parser is based on that of `tulip`
from __future__ import absolute_import
from omega.logic.ast import Nodes
import astutils


TABMODULE = 'omega.logic.ltl_parsetab'


class Lexer(astutils.Lexer):
    """Token rules to build LTL lexer."""

    reserved = {
        'VARIABLE': 'VARIABLE',
        'VARIABLES': 'VARIABLES',
        'CONSTANT': 'CONSTANT',
        'CONSTANTS': 'CONSTANTS',
        'ite': 'ITE',
        'X': 'NEXT',
        'FALSE': 'FALSE',
        'False': 'FALSE',
        'false': 'FALSE',
        'TRUE': 'TRUE',
        'True': 'TRUE',
        'true': 'TRUE',
        'LET': 'LET',
        'IN': 'IN_EXPR',
        'IF': 'IF',
        'THEN': 'THEN',
        'ELSE': 'ELSE',
        'U': 'UNTIL',
        'W': 'WEAK_UNTIL',
        'V': 'RELEASE',
        'S': 'SINCE',  # as in NuSMV
        'T': 'TRIGGER'}
    values = {'next': 'X'}
    delimiters = ['LPAREN', 'RPAREN', 'DQUOTES', 'COMMA']
    operators = [
        'NOT', 'AND', 'OR', 'XOR', 'IMPLIES', 'EQUIV',
        'EQUALS', 'NEQUALS', 'LT', 'LE', 'GT', 'GE',
        'PLUS', 'MINUS', 'TIMES', 'DIV', 'MOD', 'TRUNCATE',
        'PREVIOUS', 'WEAK_PREVIOUS', 'HISTORICALLY',
        'ALWAYS', 'EVENTUALLY',
        'ONCE', 'PRIME', 'DOTS', 'AT',
        'FORALL', 'EXISTS', 'RENAME', 'IN',
        'COLON', 'DEF']
    misc = ['NAME', 'NUMBER']

    def t_NAME(self, t):
        r'[A-Za-z_][A-za-z0-9_]*'
        t.value = self.values.get(t.value, t.value)
        t.type = self.reserved.get(t.value, 'NAME')
        return t

    def t_AND(self, t):
        r'\&\&|\&|/\\'
        t.value = '/\\'
        return t

    def t_OR(self, t):
        r'\|\||\||\\/'
        t.value = r'\/'
        return t

    def t_NEQUALS(self, t):
        r'\!\=|/\='
        # symbol `#` already marks comments
        return t

    def t_NOT(self, t):
        r'~|\!'
        t.value = '~'
        return t

    def t_IMPLIES(self, t):
        r'\=>|->'
        t.value = '=>'
        return t

    def t_EQUIV(self, t):
        r'\<\=>|\<->'
        t.value = '<=>'
        return t

    t_DEF = r'\=\='
    # quantifiers
    t_FORALL = r'\\A'
    t_EXISTS = r'\\E'
    t_RENAME = r'\\S'  # for arbitrary substitution (compose)
    # conjoin and quantify existentially
    t_COLON = r'\:'
    # set theory
    t_IN = r'\\in'
    # Boolean
    t_XOR = r'\^'
    # comparators
    t_EQUALS = r'\='
    t_LT = r'\<'
    t_LE = r'\<\='
    t_GT = r'>\='
    t_GE = r'>'
    # delimiters
    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_COMMA = r','
    t_DQUOTES = r'\"'
    # arithmetic
    t_PLUS = r'\+'
    t_MINUS = r'-'
    t_TIMES = r'\*'
    t_DIV = r'/'
    t_MOD = r'\%'
    t_TRUNCATE = r'<<>>'
    # temporal
    t_PREVIOUS = r'--X'
    t_WEAK_PREVIOUS = r'-X'
    t_HISTORICALLY = r'-\[\]'
    t_ONCE = r'-\<\>'
    t_ALWAYS = r'\[\]'
    t_EVENTUALLY = r'\<\>'
    t_DOTS = r'\.\.'
    t_PRIME = r"\'"
    # other
    t_AT = r'@'
    t_NUMBER = r'\d+'
    t_ignore = " \t"

    def t_comment(self, t):
        r'\#.*'
        return

    def t_multiline_comment(self, t):
        r'\(\*[\s\S]*?\*\)'
        return

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")


class Parser(astutils.Parser):
    """Production rules to build LTL parser."""

    tabmodule = TABMODULE
    start = 'start'
    # lowest to highest
    # based on precedence in `spin.y`
    precedence = (
        ('nonassoc', 'REDUCE_LIST'),
        ('nonassoc', 'DEF'),
        ('nonassoc', 'LET_IN', 'IF_THEN_ELSE', 'CONJ_LIST'),
        ('left', 'COLON'),
        ('left', 'EQUIV'),
        ('left', 'IMPLIES'),
        ('left', 'XOR'),
        ('left', 'OR'),
        ('left', 'AND'),
        ('left', 'ALWAYS', 'EVENTUALLY', 'HISTORICALLY', 'ONCE'),
        ('left', 'UNTIL', 'WEAK_UNTIL', 'RELEASE', 'SINCE', 'TRIGGER'),
        ('left', 'EQUALS', 'NEQUALS'),
        ('left', 'LT', 'LE', 'GT', 'GE', 'IN'),
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES', 'DIV', 'MOD'),
        ('right', 'NOT', 'UMINUS'),
        ('left', 'RENAME'),
        ('left', 'FORALL', 'EXISTS'),
        ('right', 'NEXT', 'WEAK_PREVIOUS', 'PREVIOUS'),
        ('nonassoc', 'DOTS'),
        ('left', 'PRIME'))
    Lexer = Lexer
    nodes = Nodes

    def p_start(self, p):
        """start : module
                 | expr
        """
        p[0] = p[1]

    def p_module(self, p):
        """module : units"""
        p[0] = p[1]

    def p_units_iter(self, p):
        """units : units unit"""
        p[1].append(p[2])
        p[0] = p[1]

    def p_units_end(self, p):
        """units : unit"""
        p[0] = [p[1]]

    def p_unit(self, p):
        """unit : def
                | var_decl
                | const_decl
        """
        p[0] = p[1]

    def p_variable_declaration(self, p):
        """var_decl : VARIABLE list
                    | VARIABLES list
        """
        p[0] = self.nodes.Operator(p[1], p[2])

    def p_constant_declaration(self, p):
        """const_decl : CONSTANT list
                      | CONSTANTS list
        """
        p[0] = self.nodes.Operator(p[1], p[2])

    def p_defs_iter(self, p):
        """defs : defs def"""
        p[1].append(p[2])
        p[0] = p[1]

    def p_defs_end(self, p):
        """defs : def"""
        p[0] = [p[1]]

    def p_operator_definition(self, p):
        """def : NAME DEF expr """
        name = p[1]
        u = self.nodes.Terminal(name, dtype='opname')
        p[0] = self.nodes.Binary('==', u, p[3])

    def p_let_in(self, p):
        """expr : LET defs IN_EXPR expr %prec LET_IN"""
        p[0] = self.nodes.Operator(p[1], p[2], p[4])

    def p_nullary(self, p):
        """expr : TRUE
                | FALSE
        """
        p[0] = self.nodes.Bool(p[1])

    def p_postfix_next(self, p):
        """expr : expr PRIME"""
        p[0] = self.nodes.Unary('X', p[1])

    def p_unary(self, p):
        """expr : NOT expr
                | ALWAYS expr
                | EVENTUALLY expr
                | NEXT expr
                | WEAK_PREVIOUS expr
                | PREVIOUS expr
                | HISTORICALLY expr
                | ONCE expr
        """
        p[0] = self.nodes.Unary(p[1], p[2])

    def p_binary_connective(self, p):
        """expr : expr AND expr
                | expr OR expr
                | expr XOR expr
                | expr IMPLIES expr
                | expr EQUIV expr
                | expr UNTIL expr
                | expr WEAK_UNTIL expr
                | expr RELEASE expr
                | expr SINCE expr
                | expr TRIGGER expr
        """
        p[0] = self.nodes.Binary(p[2], p[1], p[3])

    def p_binary_predicate(self, p):
        """expr : expr EQUALS expr
                | expr NEQUALS expr
                | expr LT expr
                | expr LE expr
                | expr GT expr
                | expr GE expr
        """
        p[0] = self.nodes.Comparator(p[2], p[1], p[3])

    def p_binary_function(self, p):
        """expr : expr TIMES expr
                | expr DIV expr
                | expr MOD expr
                | expr PLUS expr
                | expr MINUS expr
        """
        p[0] = self.nodes.Arithmetic(p[2], p[1], p[3])

    # CAUTION: indentation ignored: use parentheses
    def p_junction_list(self, p):
        """expr : junc_list  %prec REDUCE_LIST"""
        p[0] = p[1]

    def p_junction_list_iter(self, p):
        """junc_list : junc_list AND expr
                     | junc_list OR expr
        """
        p[0] = self.nodes.Binary(p[2], p[1], p[3])

    def p_conjunction_list_end(self, p):
        """junc_list : AND expr
                     | OR expr  %prec CONJ_LIST
        """
        p[0] = p[2]

    def p_truncator(self, p):
        """expr : expr TRUNCATE number"""
        # keep separate to allow overriding
        p[0] = self.nodes.Arithmetic(p[2], p[1], p[3])

    def p_if_then_else(self, p):
        """expr : IF expr THEN expr ELSE expr %prec IF_THEN_ELSE"""
        p[0] = self.nodes.Operator('ite', p[2], p[4], p[6])

    def p_ternary_conditional(self, p):
        """expr : ITE LPAREN expr COMMA expr COMMA expr RPAREN"""
        p[0] = self.nodes.Operator(p[1], p[3], p[5], p[7])

    def p_quantifier(self, p):
        """expr : FORALL list COLON expr
                | EXISTS list COLON expr
        """
        p[0] = self.nodes.Operator(p[1], p[2], p[4])

    def p_substitute(self, p):
        """expr : RENAME pairs COLON expr"""
        p[0] = self.nodes.Operator(p[1], p[2], p[4])

    def p_in_range(self, p):
        """expr : expr IN expr"""
        p[0] = self.nodes.Binary(p[2], p[1], p[3])

    def p_range(self, p):
        """expr : number DOTS number"""
        p[0] = self.nodes.Binary(p[2], p[1], p[3])

    def p_varlist_iter(self, p):
        """list : list COMMA expr"""
        p[1].append(p[3])
        p[0] = p[1]

    def p_varlist_end(self, p):
        """list : expr"""
        p[0] = [p[1]]

    def p_pairs_iter(self, p):
        """pairs : pairs COMMA pair"""
        p[1].append(p[3])
        p[0] = p[1]

    def p_pairs_end(self, p):
        """pairs : pair"""
        p[0] = [p[1]]

    def p_pair(self, p):
        """pair : expr DIV expr"""
        p[0] = (p[1], p[3])

    def p_paren(self, p):
        """expr : LPAREN expr RPAREN"""
        p[0] = p[2]

    def p_var(self, p):
        """expr : var"""
        p[0] = p[1]

    def p_var_unprimed(self, p):
        """var : NAME"""
        p[0] = self.nodes.Var(p[1])

    def p_bdd_node(self, p):
        """expr : AT number"""
        p[0] = self.nodes.Operator(p[1], p[2])

    def p_number_expr(self, p):
        """expr : number"""
        p[0] = p[1]

    def p_number(self, p):
        """number : NUMBER"""
        p[0] = self.nodes.Num(p[1])

    def p_negative_number(self, p):
        """number : MINUS NUMBER %prec UMINUS"""
        p[0] = self.nodes.Num('-' + p[2])

    def p_string(self, p):
        """expr : DQUOTES NAME DQUOTES"""
        p[0] = self.nodes.Str('"' + p[2] + '"')


def _rewrite_tables(outputdir='./'):
    astutils.rewrite_tables(Parser, TABMODULE, outputdir)


if __name__ == '__main__':
    import logging
    log = logging.getLogger('astutils')
    log.setLevel(logging.DEBUG)
    log.addHandler(logging.StreamHandler())
    _rewrite_tables()



"""Test `omega.symbolic.enumeration."""
import logging

import networkx as nx

from omega.symbolic import enumeration as enum
from omega.symbolic import symbolic


logging.getLogger('astutils').setLevel('ERROR')
logging.getLogger('omega').setLevel('ERROR')


def test_relation_to_graph():
    a = symbolic.Automaton()
    a.vars['x'] = dict(type='int', dom=(0, 5), owner='sys')
    a.vars['y'] = dict(type='bool', owner='sys')
    aut = a.build()
    u = aut.add_expr("(x = 4) & y & (x' = 0)")
    care_bits = aut.bdd.vars
    g = enum.relation_to_graph(
        u, aut, care_bits=care_bits)
    assert len(g) == 3, g.nodes()
    keys = ('x', 'y')
    r = {tuple(d[k] for k in keys)
         for k, d in g.nodes(data=True)}
    r_ = {(4, True), (0, False), (0, True)}
    assert r == r_, r


def test_enumerate_int():
    b = [0, 0, 0]
    k = _enum_int(b)
    assert k == 0, k
    b = [0, 1, 0, 0]
    k = _enum_int(b)
    assert k == 2, k
    b = [1, 0, 1, 0]
    k = _enum_int(b)
    assert k == 5, k
    b = [0, 1, 0, 1]
    k = _enum_int(b)
    assert k == -6, k
    # partial assignment
    b = [0, 1, None, 0]
    r = set(enum._enumerate_int(b))
    assert r == {2, 6}, r
    b = [None, 1, None, 0]
    r = set(enum._enumerate_int(b))
    assert r == {2, 3, 6, 7}, r


def _enum_int(b):
    gen = enum._enumerate_int(b)
    (k,) = list(gen)
    return k


def test_product_iter():
    sets = dict(x={1, 2}, y={3, 4})
    model = dict(z=True)
    r = list(enum._take_product_iter(sets, model))
    r_ = [dict(x=1, y=3, z=True),
          dict(x=2, y=3, z=True),
          dict(x=1, y=4, z=True),
          dict(x=2, y=4, z=True)]
    for d in r:
        assert d in r_, d
    for d in r_:
        assert d in r, d
    # trivial case
    sets = dict()
    r = list(enum._take_product_iter(sets, model))
    assert r == [model], r


def test_unprime_any_primed():
    model = {"x'": 5, "y": False, "w'": 6}
    r = enum._unprime_any_primed(model)
    r_ = dict(x=5, y=False, w=6)
    assert r == r_, r


def test_find_or_add_model():
    model = dict(x=5, y=False)
    umap = dict()
    keys = ('x', 'y')
    u = enum._find_or_add_model(model, umap, keys)
    k = (5, False)
    assert k in umap, umap
    u_ = umap[k]
    assert u == u_, (u, u_)


def test_format_nx():
    g = nx.DiGraph()
    sorted_vars = ['x', 'y', 'w']
    g.add_node(0, x=5, y=3)
    g.add_node(1, w=True)
    g.add_edge(1, 0)
    h, umap = enum._format_nx(g, sorted_vars)
    assert len(h) == len(g), (h.nodes(), g.nodes())
    assert len(h) == len(umap), (h.nodes(), umap)
    ux = r' (x=5) &and; (y=3) \l'
    uy = r' (w=True) \l'
    assert h.has_edge(uy, ux), h.edges()


def test_square_conj():
    # single line
    c = ['a', 'b']
    s = enum._square_conj(c, 2, op='/\\')
    s_ = r' (a) /\ (b) \l'
    assert s == s_, s
    # multi-line conjunction
    c = ['a', 'b', 'c']
    s = enum._square_conj(c, 3, op='/\\')
    s_ = r'/\ (a) /\ (b) \l/\ (c) '
    assert s == s_, s
